@@ -50,21 +50,19 @@ void _glfwInputWindowFocus(_GLFWwindow* window, bool focused)
 
     if (!focused)
     {
-        int key, button;
         _glfw.focusedWindowId = 0;
 
-        for (key = 0;  key <= GLFW_KEY_LAST;  key++)
+        for (unsigned i = 0; i < arraysz(window->activated_keys); i++)
         {
-            if (window->keys[key] == GLFW_PRESS)
+            if (window->activated_keys[i].key > 0 && window->activated_keys[i].action == GLFW_PRESS)
             {
-                const int native_key = _glfwPlatformGetNativeKeyForKey(key);
-                GLFWkeyevent ev;
-                _glfwInitializeKeyEvent(&ev, key, native_key, GLFW_RELEASE, 0);
+                const int native_key = _glfwPlatformGetNativeKeyForKey(window->activated_keys[i].key);
+                GLFWkeyevent ev = {.key = window->activated_keys[i].key, .native_key = native_key, .action = GLFW_RELEASE};
                 _glfwInputKeyboard(window, &ev);
             }
         }
 
-        for (button = 0;  button <= GLFW_MOUSE_BUTTON_LAST;  button++)
+        for (int button = 0;  button <= GLFW_MOUSE_BUTTON_LAST;  button++)
         {
             if (window->mouseButtons[button] == GLFW_PRESS)
                 _glfwInputMouseClick(window, button, GLFW_RELEASE, 0);
@@ -703,8 +701,8 @@ GLFWAPI void glfwSetWindowSizeIncrements(GLFWwindow* handle, int widthincr, int
 {
     _GLFWwindow* window = (_GLFWwindow*) handle;
     assert(window != NULL);
-    assert(widthincr >= 0);
-    assert(heightincr >= 0);
+    assert(widthincr >= 0 || widthincr == GLFW_DONT_CARE);
+    assert(heightincr >= 0 || heightincr == GLFW_DONT_CARE);
 
     _GLFW_REQUIRE_INIT();
 
@@ -1060,6 +1058,10 @@ GLFWAPI bool glfwToggleFullscreen(GLFWwindow* wh, unsigned int flags) {
     return false;
 }
 
+GLFWAPI bool glfwIsFullscreen(GLFWwindow* wh, unsigned int flags) {
+    return _glfwPlatformIsFullscreen((_GLFWwindow*)wh, flags);
+}
+
 GLFWAPI void glfwSetWindowUserPointer(GLFWwindow* handle, void* pointer)
 {
     _GLFWwindow* window = (_GLFWwindow*) handle;
