@@ -1,34 +1,35 @@
 #!/usr/bin/env python3
-# vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2017, Kovid Goyal <kovid at kovidgoyal.net>
 
 import contextlib
 import os
 import re
-import socket
 import signal
+import socket
 import sys
 import zlib
 from base64 import standard_b64encode
-from functools import lru_cache
 from math import ceil
 from tempfile import NamedTemporaryFile
 from typing import (
     Dict, Generator, List, NamedTuple, Optional, Pattern, Tuple, Union
 )
 
-from kitty.guess_mime_type import guess_type
 from kitty.cli import parse_args
 from kitty.cli_stub import IcatCLIOptions
 from kitty.constants import appname
+from kitty.guess_mime_type import guess_type
+from kitty.rgb import to_color
+from kitty.types import run_once
 from kitty.typing import GRT_f, GRT_t
 from kitty.utils import (
-    TTYIO, ScreenSize, ScreenSizeGetter, fit_image, screen_size_function
+    TTYIO, ScreenSize, ScreenSizeGetter, screen_size_function
 )
 
 from ..tui.images import (
-    ConvertFailed, GraphicsCommand, NoImageMagick, OpenFailed, convert, fsenc,
-    identify
+    ConvertFailed, Dispose, GraphicsCommand, NoImageMagick, OpenFailed,
+    OutdatedImageMagick, RenderedImage, fsenc, identify,
+    render_as_single_image, render_image
 )
 from ..tui.operations import clear_images_on_screen, raw_mode
 
@@ -55,6 +56,19 @@ are smaller than the specified area to be scaled up to use as much
 of the specified area as possible.
 
 
+--background
+default=none
+Specify a background color, this will cause transparent images to be composited on
+top of the specified color.
+
+
+--mirror
+default=none
+type=choices
+choices=none,horizontal,vertical,both
+Mirror the image about a horizontal or vertical axis or both.
+
+
 --clear
 type=bool-set
 Remove all images currently displayed on the screen.
@@ -88,7 +102,7 @@ detecting image display support.
 --print-window-size
 type=bool-set
 Print out the window size as :italic:`widthxheight` (in pixels) and quit. This is a
-convenience method to query the window size if using kitty icat from a
+convenience method to query the window size if using :code:`kitty +kitten icat` from a
 scripting language that cannot make termios calls.
 
 
@@ -102,19 +116,27 @@ but you can turn it off or on explicitly, if needed.
 
 --silent
 type=bool-set
-Do not print out anything to stdout during operation.
+Do not print out anything to STDOUT during operation.
 
 
 --z-index -z
 default=0
 Z-index of the image. When negative, text will be displayed on top of the image. Use
 a double minus for values under the threshold for drawing images under cell background
-colors. For example, --1 evaluates as -1,073,741,825.
+colors. For example, :code:`--1` evaluates as -1,073,741,825.
+
+
+--loop -l
+default=-1
+type=int
+Number of times to loop animations. Negative values loop forever. Zero means
+only the first frame of the animation is displayed. Otherwise, the animation
+is looped the specified number of times.
 
 
 --hold
 type=bool-set
-Wait for keypress before exiting after displaying the images.
+Wait for a key press before exiting after displaying the images.
 '''
 
 
@@ -134,9 +156,9 @@ def get_screen_size() -> ScreenSize:
     return screen_size()
 
 
-@lru_cache(maxsize=2)
+@run_once
 def options_spec() -> str:
-    return OPTIONS.format(appname='{}-icat'.format(appname))
+    return OPTIONS.format(appname=f'{appname}-icat')
 
 
 def write_gr_cmd(cmd: GraphicsCommand, payload: Optional[bytes] = None) -> None:
@@ -159,20 +181,14 @@ def set_cursor(cmd: GraphicsCommand, width: int, height: int, align: str) -> Non
     ss = get_screen_size()
     cw = int(ss.width / ss.cols)
     num_of_cells_needed = int(ceil(width / cw))
-    if num_of_cells_needed > ss.cols:
-        w, h = fit_image(width, height, ss.width, height)
-        ch = int(ss.height / ss.rows)
-        num_of_rows_needed = int(ceil(height / ch))
-        cmd.c, cmd.r = ss.cols, num_of_rows_needed
-    else:
-        cmd.X = calculate_in_cell_x_offset(width, cw, align)
-        extra_cells = 0
-        if align == 'center':
-            extra_cells = (ss.cols - num_of_cells_needed) // 2
-        elif align == 'right':
-            extra_cells = (ss.cols - num_of_cells_needed)
-        if extra_cells:
-            sys.stdout.buffer.write(b' ' * extra_cells)
+    cmd.X = calculate_in_cell_x_offset(width, cw, align)
+    extra_cells = 0
+    if align == 'center':
+        extra_cells = (ss.cols - num_of_cells_needed) // 2
+    elif align == 'right':
+        extra_cells = (ss.cols - num_of_cells_needed)
+    if extra_cells:
+        sys.stdout.buffer.write(b' ' * extra_cells)
 
 
 def set_cursor_for_place(place: 'Place', cmd: GraphicsCommand, width: int, height: int, align: str) -> None:
@@ -186,19 +202,24 @@ def set_cursor_for_place(place: 'Place', cmd: GraphicsCommand, width: int, heigh
         extra_cells = (place.width - num_of_cells_needed) // 2
     elif align == 'right':
         extra_cells = place.width - num_of_cells_needed
-    sys.stdout.buffer.write('\033[{};{}H'.format(place.top + 1, x + extra_cells).encode('ascii'))
+    sys.stdout.buffer.write(f'\033[{place.top + 1};{x + extra_cells}H'.encode('ascii'))
 
 
 def write_chunked(cmd: GraphicsCommand, data: bytes) -> None:
+    cmd = cmd.clone()
     if cmd.f != 100:
         data = zlib.compress(data)
         cmd.o = 'z'
     data = standard_b64encode(data)
+    ac = cmd.a
+    quiet = cmd.q
     while data:
         chunk, data = data[:4096], data[4096:]
         cmd.m = 1 if data else 0
         write_gr_cmd(cmd, chunk)
         cmd.clear()
+        cmd.a = ac
+        cmd.q = quiet
 
 
 def show(
@@ -207,7 +228,8 @@ def show(
     fmt: 'GRT_f',
     transmit_mode: 'GRT_t' = 't',
     align: str = 'center',
-    place: Optional['Place'] = None
+    place: Optional['Place'] = None,
+    use_number: int = 0
 ) -> None:
     cmd = GraphicsCommand()
     cmd.a = 'T'
@@ -215,6 +237,9 @@ def show(
     cmd.s = width
     cmd.v = height
     cmd.z = zindex
+    if use_number:
+        cmd.I = use_number  # noqa
+        cmd.q = 2
     if place:
         set_cursor_for_place(place, cmd, width, height, align)
     else:
@@ -232,6 +257,58 @@ def show(
         write_chunked(cmd, data)
 
 
+def show_frames(frame_data: RenderedImage, use_number: int, loops: int) -> None:
+    transmit_cmd = GraphicsCommand()
+    transmit_cmd.a = 'f'
+    transmit_cmd.I = use_number  # noqa
+    transmit_cmd.q = 2
+    if can_transfer_with_files:
+        transmit_cmd.t = 't'
+    transmit_cmd.f = 24 if frame_data.mode == 'rgb' else 32
+
+    def control(frame_number: int = 0, loops: Optional[int] = None, gap: Optional[int] = 0, animation_control: int = 0) -> None:
+        cmd = GraphicsCommand()
+        cmd.a = 'a'
+        cmd.I = use_number  # noqa
+        cmd.r = frame_number
+        if loops is not None:
+            cmd.v = loops + 1
+        if gap is not None:
+            cmd.z = gap if gap > 0 else -1
+        if animation_control:
+            cmd.s = animation_control
+        write_gr_cmd(cmd)
+
+    anchor_frame = 0
+
+    for frame in frame_data.frames:
+        frame_number = frame.index + 1
+        if frame.dispose < Dispose.previous:
+            anchor_frame = frame_number
+        if frame_number == 1:
+            control(frame_number, gap=frame.gap, loops=None if loops < 1 else loops)
+            continue
+        if frame.dispose is Dispose.previous:
+            if anchor_frame != frame_number:
+                transmit_cmd.c = anchor_frame
+        else:
+            transmit_cmd.c = (frame_number - 1) if frame.needs_blend else 0
+        transmit_cmd.s = frame.width
+        transmit_cmd.v = frame.height
+        transmit_cmd.x = frame.canvas_x
+        transmit_cmd.y = frame.canvas_y
+        transmit_cmd.z = frame.gap if frame.gap > 0 else -1
+        if can_transfer_with_files:
+            write_gr_cmd(transmit_cmd, standard_b64encode(os.path.abspath(frame.path).encode(fsenc)))
+        else:
+            with open(frame.path, 'rb') as f:
+                data = f.read()
+            write_chunked(transmit_cmd, data)
+        if frame_number == 2:
+            control(animation_control=2)
+    control(animation_control=3)
+
+
 def parse_z_index(val: str) -> int:
     origin = 0
     if val.startswith('--'):
@@ -244,6 +321,9 @@ class ParsedOpts:
 
     place: Optional['Place'] = None
     z_index: int = 0
+    remove_alpha: str = ''
+    flip: bool = False
+    flop: bool = False
 
 
 def process(path: str, args: IcatCLIOptions, parsed_opts: ParsedOpts, is_tempfile: bool) -> bool:
@@ -253,8 +333,10 @@ def process(path: str, args: IcatCLIOptions, parsed_opts: ParsedOpts, is_tempfil
     available_height = parsed_opts.place.height * (ss.height // ss.rows) if parsed_opts.place else 10 * m.height
     needs_scaling = m.width > available_width or m.height > available_height
     needs_scaling = needs_scaling or args.scale_up
+    needs_conversion = needs_scaling or bool(parsed_opts.remove_alpha) or parsed_opts.flip or parsed_opts.flop
     file_removed = False
-    if m.fmt == 'png' and not needs_scaling:
+    use_number = 0
+    if m.fmt == 'png' and not needs_conversion:
         outfile = path
         transmit_mode: 'GRT_t' = 't' if is_tempfile else 'f'
         fmt: 'GRT_f' = 100
@@ -263,8 +345,29 @@ def process(path: str, args: IcatCLIOptions, parsed_opts: ParsedOpts, is_tempfil
     else:
         fmt = 24 if m.mode == 'rgb' else 32
         transmit_mode = 't'
-        outfile, width, height = convert(path, m, available_width, available_height, args.scale_up)
-    show(outfile, width, height, parsed_opts.z_index, fmt, transmit_mode, align=args.align, place=parsed_opts.place)
+        if len(m) == 1 or args.loop == 0:
+            outfile, width, height = render_as_single_image(
+                path, m, available_width, available_height, args.scale_up,
+                remove_alpha=parsed_opts.remove_alpha, flip=parsed_opts.flip, flop=parsed_opts.flop)
+        else:
+            import struct
+            use_number = max(1, struct.unpack('@I', os.urandom(4))[0])
+            with NamedTemporaryFile() as f:
+                prefix = f.name
+            frame_data = render_image(
+                path, prefix, m, available_width, available_height, args.scale_up,
+                remove_alpha=parsed_opts.remove_alpha, flip=parsed_opts.flip, flop=parsed_opts.flop)
+            outfile, width, height = frame_data.frames[0].path, frame_data.width, frame_data.height
+    show(
+        outfile, width, height, parsed_opts.z_index, fmt, transmit_mode,
+        align=args.align, place=parsed_opts.place, use_number=use_number
+    )
+    if use_number:
+        show_frames(frame_data, use_number, args.loop)
+        if not can_transfer_with_files:
+            for fr in frame_data.frames:
+                with contextlib.suppress(FileNotFoundError):
+                    os.unlink(fr.path)
     if not args.place:
         print()  # ensure cursor is on a new line
     return file_removed
@@ -281,7 +384,7 @@ def scan(d: str) -> Generator[Tuple[str, str], None, None]:
 def detect_support(wait_for: float = 10, silent: bool = False) -> bool:
     global can_transfer_with_files
     if not silent:
-        print('Checking for graphics ({}s max. wait)...'.format(wait_for), end='\r')
+        print(f'Checking for graphics ({wait_for}s max. wait)...', end='\r')
     sys.stdout.flush()
     try:
         received = b''
@@ -302,7 +405,8 @@ def detect_support(wait_for: float = 10, silent: bool = False) -> bool:
             return 1 not in responses or 2 not in responses
 
         with NamedTemporaryFile() as f:
-            f.write(b'abcd'), f.flush()
+            f.write(b'abcd')
+            f.flush()
             gc = GraphicsCommand()
             gc.a = 'q'
             gc.s = gc.v = gc.i = 1
@@ -360,15 +464,15 @@ def process_single_item(
     item: Union[bytes, str],
     args: IcatCLIOptions,
     parsed_opts: ParsedOpts,
-    url_pat: Optional[Pattern] = None,
+    url_pat: Optional['Pattern[str]'] = None,
     maybe_dir: bool = True
 ) -> None:
     is_tempfile = False
     file_removed = False
     try:
         if isinstance(item, bytes):
-            tf = NamedTemporaryFile(prefix='stdin-image-data-', delete=False)
-            tf.write(item), tf.close()
+            with NamedTemporaryFile(prefix='stdin-image-data-', delete=False) as tf:
+                tf.write(item)
             item = tf.name
             is_tempfile = True
         if url_pat is not None and url_pat.match(item) is not None:
@@ -378,7 +482,7 @@ def process_single_item(
                     with socket_timeout(30):
                         urlretrieve(item, filename=tf.name)
                 except Exception as e:
-                    raise SystemExit('Failed to download image at URL: {} with error: {}'.format(item, e))
+                    raise SystemExit(f'Failed to download image at URL: {item} with error: {e}')
                 item = tf.name
             is_tempfile = True
             file_removed = process(item, args, parsed_opts, is_tempfile)
@@ -405,25 +509,25 @@ def process_single_item(
 
 def main(args: List[str] = sys.argv) -> None:
     global can_transfer_with_files
-    cli_opts, items_ = parse_args(args[1:], options_spec, usage, help_text, '{} +kitten icat'.format(appname), result_class=IcatCLIOptions)
+    cli_opts, items_ = parse_args(args[1:], options_spec, usage, help_text, f'{appname} +kitten icat', result_class=IcatCLIOptions)
     items: List[Union[str, bytes]] = list(items_)
 
     if cli_opts.print_window_size:
         screen_size_function.cache_clear()
         with open(os.ctermid()) as tty:
             ss = screen_size_function(tty)()
-        print('{}x{}'.format(ss.width, ss.height), end='')
+        print(f'{ss.width}x{ss.height}', end='')
         raise SystemExit(0)
 
     if not sys.stdout.isatty():
         sys.stdout = open(os.ctermid(), 'w')
     stdin_data = None
-    if cli_opts.stdin == 'yes' or (not sys.stdin.isatty() and cli_opts.stdin == 'detect'):
+    if cli_opts.stdin == 'yes' or (cli_opts.stdin == 'detect' and sys.stdin is not None and not sys.stdin.isatty()):
         stdin_data = sys.stdin.buffer.read()
         if stdin_data:
             items.insert(0, stdin_data)
         sys.stdin.close()
-        sys.stdin = open(os.ctermid(), 'r')
+        sys.stdin = open(os.ctermid())
 
     screen_size = get_screen_size_function()
     signal.signal(signal.SIGWINCH, lambda signum, frame: setattr(screen_size, 'changed', True))
@@ -438,12 +542,19 @@ def main(args: List[str] = sys.argv) -> None:
         try:
             parsed_opts.place = parse_place(cli_opts.place)
         except Exception:
-            raise SystemExit('Not a valid place specification: {}'.format(cli_opts.place))
+            raise SystemExit(f'Not a valid place specification: {cli_opts.place}')
 
     try:
         parsed_opts.z_index = parse_z_index(cli_opts.z_index)
     except Exception:
-        raise SystemExit('Not a valid z-index specification: {}'.format(cli_opts.z_index))
+        raise SystemExit(f'Not a valid z-index specification: {cli_opts.z_index}')
+    if cli_opts.background != 'none':
+        ra = to_color(cli_opts.background)
+        if ra is None:
+            raise SystemExit(f'Not a valid color specification: {cli_opts.background}')
+        parsed_opts.remove_alpha = ra.as_sharp
+    parsed_opts.flip = cli_opts.mirror in ('both', 'vertical')
+    parsed_opts.flop = cli_opts.mirror in ('both', 'horizontal')
 
     if cli_opts.detect_support:
         if not detect_support(wait_for=cli_opts.detection_timeout, silent=True):
@@ -467,13 +578,26 @@ def main(args: List[str] = sys.argv) -> None:
             raise SystemExit(f'The --place option can only be used with a single image, not {items}')
         sys.stdout.buffer.write(b'\0337')  # save cursor
     url_pat = re.compile(r'(?:https?|ftp)://', flags=re.I)
+
+    def hold_if_needed(exit_code_or_msg: Union[int, str]) -> None:
+        if cli_opts.hold:
+            if isinstance(exit_code_or_msg, str):
+                print(exit_code_or_msg, file=sys.stderr, flush=True)
+                exit_code_or_msg = 1
+            with open(os.ctermid()) as tty, raw_mode(tty.fileno()):
+                tty.buffer.read(1)
+        raise SystemExit(exit_code_or_msg)
+
     for item in items:
         try:
             process_single_item(item, cli_opts, parsed_opts, url_pat)
         except NoImageMagick as e:
-            raise SystemExit(str(e))
+            hold_if_needed(str(e))
+        except OutdatedImageMagick as e:
+            print(e.detailed_error, file=sys.stderr)
+            hold_if_needed(str(e))
         except ConvertFailed as e:
-            raise SystemExit(str(e))
+            hold_if_needed(str(e))
         except OpenFailed as e:
             errors.append(e)
     if parsed_opts.place:
@@ -481,11 +605,8 @@ def main(args: List[str] = sys.argv) -> None:
     if errors:
         for err in errors:
             print(err, file=sys.stderr)
-    if cli_opts.hold:
-        with open(os.ctermid()) as tty:
-            with raw_mode(tty.fileno()):
-                tty.buffer.read(1)
-    raise SystemExit(1 if errors else 0)
+    hold_if_needed(1 if errors else 0)
+    raise SystemExit()
 
 
 if __name__ == '__main__':
