@@ -7,7 +7,9 @@
 
 
 #include "state.h"
+#include "cleanup.h"
 #include "monotonic.h"
+#include <Carbon/Carbon.h>
 #include <Cocoa/Cocoa.h>
 #ifndef KITTY_USE_DEPRECATED_MACOS_NOTIFICATION_API
 #include <UserNotifications/UserNotifications.h>
@@ -17,14 +19,22 @@
 // Needed for _NSGetProgname
 #include <crt_externs.h>
 #include <objc/runtime.h>
-#include <xlocale.h>
 
+#if (MAC_OS_X_VERSION_MAX_ALLOWED < 101300)
+#define NSControlStateValueOn NSOnState
+#define NSControlStateValueOff NSOffState
+#define NSControlStateValueMixed NSMixedState
+#endif
 #if (MAC_OS_X_VERSION_MAX_ALLOWED < 101200)
 #define NSWindowStyleMaskResizable NSResizableWindowMask
 #define NSEventModifierFlagOption NSAlternateKeyMask
 #define NSEventModifierFlagCommand NSCommandKeyMask
 #define NSEventModifierFlagControl NSControlKeyMask
 #endif
+#if (MAC_OS_X_VERSION_MAX_ALLOWED < 110000)
+#define UNNotificationPresentationOptionList (1 << 3)
+#define UNNotificationPresentationOptionBanner (1 << 4)
+#endif
 
 typedef int CGSConnectionID;
 typedef int CGSWindowID;
@@ -71,52 +81,222 @@ find_app_name(void) {
     return @"kitty";
 }
 
+#define debug_key(...) if (OPT(debug_keyboard)) { fprintf(stderr, __VA_ARGS__); fflush(stderr); }
+
+// SecureKeyboardEntryController {{{
+@interface SecureKeyboardEntryController : NSObject
+
+@property (nonatomic, readonly) BOOL isDesired;
+@property (nonatomic, readonly, getter=isEnabled) BOOL enabled;
+
++ (instancetype)sharedInstance;
+
+- (void)toggle;
+- (void)update;
+
+@end
+
+@implementation SecureKeyboardEntryController {
+    int _count;
+    BOOL _desired;
+}
+
++ (instancetype)sharedInstance {
+    static id instance;
+    static dispatch_once_t onceToken;
+    dispatch_once(&onceToken, ^{
+        instance = [[self alloc] init];
+    });
+    return instance;
+}
+
+- (instancetype)init {
+    self = [super init];
+    if (self) {
+        _desired = false;
+
+        [[NSNotificationCenter defaultCenter] addObserver:self
+                                                 selector:@selector(applicationDidResignActive:)
+                                                     name:NSApplicationDidResignActiveNotification
+                                                   object:nil];
+        [[NSNotificationCenter defaultCenter] addObserver:self
+                                                 selector:@selector(applicationDidBecomeActive:)
+                                                     name:NSApplicationDidBecomeActiveNotification
+                                                   object:nil];
+        if ([NSApp isActive]) {
+            [self update];
+        }
+    }
+    return self;
+}
+
+#pragma mark - API
+
+- (void)toggle {
+    // Set _desired to the opposite of the current state.
+    _desired = !_desired;
+    debug_key("SecureKeyboardEntry: toggle called. Setting desired to %d ", _desired);
+
+    // Try to set the system's state of secure input to the desired state.
+    [self update];
+}
+
+- (BOOL)isEnabled {
+    return !!IsSecureEventInputEnabled();
+}
+
+- (BOOL)isDesired {
+    return _desired;
+}
+
+#pragma mark - Notifications
+
+- (void)applicationDidResignActive:(NSNotification *)notification {
+    (void)notification;
+    if (_count > 0) {
+        debug_key("SecureKeyboardEntry: Application resigning active.");
+        [self update];
+    }
+}
+
+- (void)applicationDidBecomeActive:(NSNotification *)notification {
+    (void)notification;
+    if (self.isDesired) {
+        debug_key("SecureKeyboardEntry: Application became active.");
+        [self update];
+    }
+}
+
+#pragma mark - Private
+
+- (BOOL)allowed {
+    return [NSApp isActive];
+}
+
+- (void)update {
+    debug_key("Update secure keyboard entry. desired=%d active=%d\n",
+         (int)self.isDesired, (int)[NSApp isActive]);
+    const BOOL secure = self.isDesired && [self allowed];
+
+    if (secure && _count > 0) {
+        debug_key("Want to turn on secure input but it's already on\n");
+        return;
+    }
+
+    if (!secure && _count == 0) {
+        debug_key("Want to turn off secure input but it's already off\n");
+        return;
+    }
+
+    debug_key("Before: IsSecureEventInputEnabled returns %d ", (int)self.isEnabled);
+    if (secure) {
+        OSErr err = EnableSecureEventInput();
+        debug_key("EnableSecureEventInput err=%d ", (int)err);
+        if (err) {
+            debug_key("EnableSecureEventInput failed with error %d ", (int)err);
+        } else {
+            _count += 1;
+        }
+    } else {
+        OSErr err = DisableSecureEventInput();
+        debug_key("DisableSecureEventInput err=%d ", (int)err);
+        if (err) {
+            debug_key("DisableSecureEventInput failed with error %d ", (int)err);
+        } else {
+            _count -= 1;
+        }
+    }
+    debug_key("After: IsSecureEventInputEnabled returns %d\n", (int)self.isEnabled);
+}
+
+@end
+// }}}
+
 @interface GlobalMenuTarget : NSObject
 + (GlobalMenuTarget *) shared_instance;
 @end
 
+#define PENDING(selector, which) - (void)selector:(id)sender { (void)sender; set_cocoa_pending_action(which, NULL); }
+
 @implementation GlobalMenuTarget
 
-- (void)show_preferences:(id)sender {
-    (void)sender;
-    set_cocoa_pending_action(PREFERENCES_WINDOW, NULL);
-}
-
-- (void)new_os_window:(id)sender {
-    (void)sender;
-    set_cocoa_pending_action(NEW_OS_WINDOW, NULL);
-}
+PENDING(edit_config_file, PREFERENCES_WINDOW)
+PENDING(new_os_window, NEW_OS_WINDOW)
+PENDING(detach_tab, DETACH_TAB)
+PENDING(close_os_window, CLOSE_OS_WINDOW)
+PENDING(close_tab, CLOSE_TAB)
+PENDING(new_tab, NEW_TAB)
+PENDING(next_tab, NEXT_TAB)
+PENDING(previous_tab, PREVIOUS_TAB)
+PENDING(new_window, NEW_WINDOW)
+PENDING(close_window, CLOSE_WINDOW)
+PENDING(reset_terminal, RESET_TERMINAL)
+PENDING(clear_terminal_and_scrollback, CLEAR_TERMINAL_AND_SCROLLBACK)
+PENDING(reload_config, RELOAD_CONFIG)
+PENDING(toggle_macos_secure_keyboard_entry, TOGGLE_MACOS_SECURE_KEYBOARD_ENTRY)
+PENDING(toggle_fullscreen, TOGGLE_FULLSCREEN)
 
 - (void)open_kitty_website_url:(id)sender {
     (void)sender;
     [[NSWorkspace sharedWorkspace] openURL:[NSURL URLWithString:@"https://sw.kovidgoyal.net/kitty/"]];
 }
 
+- (BOOL)validateMenuItem:(NSMenuItem *)item {
+    if (item.action == @selector(toggle_macos_secure_keyboard_entry:)) {
+        item.state = [SecureKeyboardEntryController sharedInstance].isDesired ? NSControlStateValueOn : NSControlStateValueOff;
+    } else if (item.action == @selector(toggle_fullscreen:)) {
+        item.title = ([NSApp currentSystemPresentationOptions] & NSApplicationPresentationFullScreen) ? @"Exit Full Screen" : @"Enter Full Screen";
+    }
+    return YES;
+}
+
+#undef PENDING
 
 + (GlobalMenuTarget *) shared_instance
 {
     static GlobalMenuTarget *sharedGlobalMenuTarget = nil;
     @synchronized(self)
     {
-        if (!sharedGlobalMenuTarget)
+        if (!sharedGlobalMenuTarget) {
             sharedGlobalMenuTarget = [[GlobalMenuTarget alloc] init];
+            SecureKeyboardEntryController *k = [SecureKeyboardEntryController sharedInstance];
+            if (!k.isDesired && [[NSUserDefaults standardUserDefaults] boolForKey:@"SecureKeyboardEntry"]) [k toggle];
+        }
         return sharedGlobalMenuTarget;
     }
 }
 
 @end
 
-static char new_window_key[32] = {0};
-static NSEventModifierFlags new_window_mods = 0;
+typedef struct {
+    char key[32];
+    NSEventModifierFlags mods;
+} GlobalShortcut;
+typedef struct {
+    GlobalShortcut new_os_window, close_os_window, close_tab, edit_config_file, reload_config;
+    GlobalShortcut previous_tab, next_tab, new_tab, new_window, close_window, reset_terminal, clear_terminal_and_scrollback;
+    GlobalShortcut toggle_macos_secure_keyboard_entry, toggle_fullscreen;
+} GlobalShortcuts;
+static GlobalShortcuts global_shortcuts;
 
 static PyObject*
-cocoa_set_new_window_trigger(PyObject *self UNUSED, PyObject *args) {
-    int mods, key;
-    if (!PyArg_ParseTuple(args, "ii", &mods, &key)) return NULL;
-    int nwm;
-    get_cocoa_key_equivalent(key, mods, new_window_key, sizeof(new_window_key), &nwm);
-    new_window_mods = nwm;
-    if (new_window_key[0]) Py_RETURN_TRUE;
+cocoa_set_global_shortcut(PyObject *self UNUSED, PyObject *args) {
+    int mods;
+    unsigned int key;
+    const char *name;
+    if (!PyArg_ParseTuple(args, "siI", &name, &mods, &key)) return NULL;
+    GlobalShortcut *gs = NULL;
+#define Q(x) if (strcmp(name, #x) == 0) gs = &global_shortcuts.x
+    Q(new_os_window); else Q(close_os_window); else Q(close_tab); else Q(edit_config_file);
+    else Q(new_tab); else Q(next_tab); else Q(previous_tab);
+    else Q(new_window); else Q(close_window); else Q(reset_terminal); else Q(clear_terminal_and_scrollback); else Q(reload_config);
+    else Q(toggle_macos_secure_keyboard_entry); else Q(toggle_fullscreen);
+#undef Q
+    if (gs == NULL) { PyErr_SetString(PyExc_KeyError, "Unknown shortcut name"); return NULL; }
+    int cocoa_mods;
+    get_cocoa_key_equivalent(key, mods, gs->key, 32, &cocoa_mods);
+    gs->mods = cocoa_mods;
+    if (gs->key[0]) Py_RETURN_TRUE;
     Py_RETURN_FALSE;
 }
 
@@ -127,7 +307,7 @@ get_dock_menu(id self UNUSED, SEL _cmd UNUSED, NSApplication *sender UNUSED) {
     if (!dockMenu) {
         GlobalMenuTarget *global_menu_target = [GlobalMenuTarget shared_instance];
         dockMenu = [[NSMenu alloc] init];
-        [[dockMenu addItemWithTitle:@"New OS window"
+        [[dockMenu addItemWithTitle:@"New OS Window"
                              action:@selector(new_os_window:)
                       keyEquivalent:@""]
                           setTarget:global_menu_target];
@@ -139,9 +319,11 @@ static PyObject *notification_activated_callback = NULL;
 
 static PyObject*
 set_notification_activated_callback(PyObject *self UNUSED, PyObject *callback) {
-    if (notification_activated_callback) Py_DECREF(notification_activated_callback);
-    notification_activated_callback = callback;
-    Py_INCREF(callback);
+    Py_CLEAR(notification_activated_callback);
+    if (callback != Py_None) {
+        notification_activated_callback = callback;
+        Py_INCREF(callback);
+    }
     Py_RETURN_NONE;
 }
 
@@ -198,6 +380,16 @@ cocoa_send_notification(PyObject *self UNUSED, PyObject *args) {
 @end
 
 @implementation NotificationDelegate
+    - (void)userNotificationCenter:(UNUserNotificationCenter *)center
+            willPresentNotification:(UNNotification *)notification
+            withCompletionHandler:(void (^)(UNNotificationPresentationOptions))completionHandler {
+        (void)(center); (void)notification;
+        UNNotificationPresentationOptions options = UNNotificationPresentationOptionSound;
+        if (@available(macOS 11.0, *)) options |= UNNotificationPresentationOptionList | UNNotificationPresentationOptionBanner;
+        else options |= (1 << 2); // UNNotificationPresentationOptionAlert avoid deprecated warning
+        completionHandler(options);
+    }
+
     - (void)userNotificationCenter:(UNUserNotificationCenter *)center
             didReceiveNotificationResponse:(UNNotificationResponse *)response
             withCompletionHandler:(void (^)(void))completionHandler {
@@ -223,6 +415,7 @@ schedule_notification(const char *identifier, const char *title, const char *bod
     if (title) content.title = @(title);
     if (body) content.body = @(body);
     if (subtitle) content.subtitle = @(subtitle);
+    content.sound = [UNNotificationSound defaultSound];
     // Deliver the notification
     static unsigned long counter = 1;
     UNNotificationRequest* request = [
@@ -282,7 +475,9 @@ cocoa_send_notification(PyObject *self UNUSED, PyObject *args) {
     if (!center.delegate) center.delegate = [[NotificationDelegate alloc] init];
     queue_notification(identifier, title, body, subtitle);
 
-    [center requestAuthorizationWithOptions:(UNAuthorizationOptionAlert)
+    // The badge permission needs to be requested as well, even though it is not used,
+    // otherwise macOS refuses to show the preference checkbox for enable/disable notification sound.
+    [center requestAuthorizationWithOptions:(UNAuthorizationOptionAlert | UNAuthorizationOptionSound | UNAuthorizationOptionBadge)
         completionHandler:^(BOOL granted, NSError * _Nullable error) {
             if (error != nil) {
                 log_error("Failed to request permission for showing notification: %s", [[error localizedDescription] UTF8String]);
@@ -328,6 +523,19 @@ cocoa_send_notification(PyObject *self UNUSED, PyObject *args) {
     return YES;
 }
 
+- (BOOL)openFileURLs:(NSPasteboard*)pasteboard
+        userData:(NSString *) UNUSED userData  error:(NSError **) UNUSED error {
+    NSDictionary *options = @{ NSPasteboardURLReadingFileURLsOnlyKey: @YES };
+    NSArray *urlArray = [pasteboard readObjectsForClasses:[NSArray arrayWithObject:[NSURL class]] options:options];
+    for (NSURL *url in urlArray) {
+        NSString *path = [url path];
+        if ([[NSFileManager defaultManager] fileExistsAtPath:path]) {
+            set_cocoa_pending_action(LAUNCH_URLS, [[[NSURL fileURLWithPath:path] absoluteString] UTF8String]);
+        }
+    }
+    return YES;
+}
+
 @end
 
 // global menu {{{
@@ -338,6 +546,12 @@ cocoa_create_global_menu(void) {
     GlobalMenuTarget *global_menu_target = [GlobalMenuTarget shared_instance];
     [NSApp setMainMenu:bar];
 
+#define MENU_ITEM(menu, title, name) { \
+    NSMenuItem *__mi = [menu addItemWithTitle:title action:@selector(name:) keyEquivalent:@(global_shortcuts.name.key)]; \
+    [__mi setKeyEquivalentModifierMask:global_shortcuts.name.mods]; \
+    [__mi setTarget:global_menu_target]; \
+}
+
     NSMenuItem* appMenuItem =
         [bar addItemWithTitle:@""
                        action:NULL
@@ -349,18 +563,17 @@ cocoa_create_global_menu(void) {
                        action:@selector(orderFrontStandardAboutPanel:)
                 keyEquivalent:@""];
     [appMenu addItem:[NSMenuItem separatorItem]];
-    [[appMenu addItemWithTitle:@"Preferences..."
-                       action:@selector(show_preferences:)
-                keyEquivalent:@","]
-                    setTarget:global_menu_target];
-
-    NSMenuItem* new_os_window_menu_item =
-        [appMenu addItemWithTitle:@"New OS window"
-                           action:@selector(new_os_window:)
-                    keyEquivalent:@(new_window_key)];
-    [new_os_window_menu_item setKeyEquivalentModifierMask:new_window_mods];
-    [new_os_window_menu_item setTarget:global_menu_target];
+    MENU_ITEM(appMenu, @"Preferences…", edit_config_file);
+    MENU_ITEM(appMenu, @"Reload Preferences", reload_config);
+    [appMenu addItem:[NSMenuItem separatorItem]];
 
+    NSMenu* servicesMenu = [[NSMenu alloc] init];
+    [NSApp setServicesMenu:servicesMenu];
+    [[appMenu addItemWithTitle:@"Services"
+                        action:NULL
+                 keyEquivalent:@""] setSubmenu:servicesMenu];
+    [servicesMenu release];
+    [appMenu addItem:[NSMenuItem separatorItem]];
 
     [appMenu addItemWithTitle:[NSString stringWithFormat:@"Hide %@", app_name]
                        action:@selector(hide:)
@@ -374,13 +587,7 @@ cocoa_create_global_menu(void) {
                 keyEquivalent:@""];
     [appMenu addItem:[NSMenuItem separatorItem]];
 
-    NSMenu* servicesMenu = [[NSMenu alloc] init];
-    [NSApp setServicesMenu:servicesMenu];
-    [[appMenu addItemWithTitle:@"Services"
-                        action:NULL
-                 keyEquivalent:@""] setSubmenu:servicesMenu];
-    [servicesMenu release];
-
+    MENU_ITEM(appMenu, @"Secure Keyboard Entry", toggle_macos_secure_keyboard_entry);
     [appMenu addItem:[NSMenuItem separatorItem]];
 
     [appMenu addItemWithTitle:[NSString stringWithFormat:@"Quit %@", app_name]
@@ -388,6 +595,24 @@ cocoa_create_global_menu(void) {
                 keyEquivalent:@"q"];
     [appMenu release];
 
+    NSMenuItem* shellMenuItem =
+        [bar addItemWithTitle:@"Shell"
+                       action:NULL
+                keyEquivalent:@""];
+    NSMenu* shellMenu = [[NSMenu alloc] initWithTitle:@"Shell"];
+    [shellMenuItem setSubmenu:shellMenu];
+    MENU_ITEM(shellMenu, @"New OS Window", new_os_window);
+    MENU_ITEM(shellMenu, @"New Tab", new_tab);
+    MENU_ITEM(shellMenu, @"New Window", new_window);
+    [shellMenu addItem:[NSMenuItem separatorItem]];
+    MENU_ITEM(shellMenu, @"Close OS Window", close_os_window);
+    MENU_ITEM(shellMenu, @"Close Tab", close_tab);
+    MENU_ITEM(shellMenu, @"Close Window", close_window);
+    [shellMenu addItem:[NSMenuItem separatorItem]];
+    MENU_ITEM(shellMenu, @"Reset", reset_terminal);
+    MENU_ITEM(shellMenu, @"Clear to Cursor Line", clear_terminal_and_scrollback);
+    [shellMenu release];
+
     NSMenuItem* windowMenuItem =
         [bar addItemWithTitle:@""
                        action:NULL
@@ -407,10 +632,14 @@ cocoa_create_global_menu(void) {
                    keyEquivalent:@""];
 
     [windowMenu addItem:[NSMenuItem separatorItem]];
-    [[windowMenu addItemWithTitle:@"Enter Full Screen"
-                           action:@selector(toggleFullScreen:)
-                    keyEquivalent:@"f"]
-     setKeyEquivalentModifierMask:NSEventModifierFlagControl | NSEventModifierFlagCommand];
+    MENU_ITEM(windowMenu, @"Show Previous Tab", previous_tab);
+    MENU_ITEM(windowMenu, @"Show Next Tab", next_tab);
+    [[windowMenu addItemWithTitle:@"Move Tab to New Window"
+                           action:@selector(detach_tab:)
+                    keyEquivalent:@""] setTarget:global_menu_target];
+
+    [windowMenu addItem:[NSMenuItem separatorItem]];
+    MENU_ITEM(windowMenu, @"Enter Full Screen", toggle_fullscreen);
     [NSApp setWindowsMenu:windowMenu];
     [windowMenu release];
 
@@ -420,10 +649,11 @@ cocoa_create_global_menu(void) {
                 keyEquivalent:@""];
     NSMenu* helpMenu = [[NSMenu alloc] initWithTitle:@"Help"];
     [helpMenuItem setSubmenu:helpMenu];
-    [[helpMenu addItemWithTitle:[NSString stringWithFormat:@"Visit %@ website", app_name]
+    [[helpMenu addItemWithTitle:[NSString stringWithFormat:@"Visit %@ Website", app_name]
                          action:@selector(open_kitty_website_url:)
                   keyEquivalent:@"?"]
                       setTarget:global_menu_target];
+    [NSApp setHelpMenu:helpMenu];
     [helpMenu release];
 
     [bar release];
@@ -436,11 +666,21 @@ cocoa_create_global_menu(void) {
 
 
     [NSApp setServicesProvider:[[[ServiceProvider alloc] init] autorelease]];
+#undef MENU_ITEM
 }
 
 void
 cocoa_update_menu_bar_title(PyObject *pytitle) {
-    NSString *title = @(PyUnicode_AsUTF8(pytitle));
+    NSString *title = nil;
+    if (OPT(macos_menubar_title_max_length) > 0 && PyUnicode_GetLength(pytitle) > OPT(macos_menubar_title_max_length)) {
+        static char fmt[64];
+        snprintf(fmt, sizeof(fmt), "%%%ld.%ldU%%s", OPT(macos_menubar_title_max_length), OPT(macos_menubar_title_max_length));
+        DECREF_AFTER_FUNCTION PyObject *st = PyUnicode_FromFormat(fmt, pytitle, "…");
+        if (st) title = @(PyUnicode_AsUTF8(st));
+    } else {
+        title = @(PyUnicode_AsUTF8(pytitle));
+    }
+    if (!title) return;
     NSMenu *bar = [NSApp mainMenu];
     if (title_menu != NULL) {
         [bar removeItem:title_menu];
@@ -479,6 +719,13 @@ cocoa_alt_option_key_pressed(NSUInteger flags) {
     return (q & flags) == q;
 }
 
+void
+cocoa_toggle_secure_keyboard_entry(void) {
+    SecureKeyboardEntryController *k = [SecureKeyboardEntryController sharedInstance];
+    [k toggle];
+    [[NSUserDefaults standardUserDefaults] setBool:k.isDesired forKey:@"SecureKeyboardEntry"];
+}
+
 void
 cocoa_focus_window(void *w) {
     NSWindow *window = (NSWindow*)w;
@@ -521,12 +768,7 @@ cocoa_get_lang(PyObject UNUSED *self) {
         locale = [[NSLocale currentLocale] localeIdentifier];
     }
     if (!locale) { Py_RETURN_NONE; }
-    // Make sure the locale value is valid, that is it can be used
-    // to construct an actual locale
     const char* locale_utf8 = [locale UTF8String];
-    locale_t test_locale = newlocale(LC_ALL_MASK, locale_utf8, NULL);
-    if (!test_locale) { Py_RETURN_NONE; }
-    freelocale(test_locale);
     return Py_BuildValue("s", locale_utf8);
 
     } // autoreleasepool
@@ -556,6 +798,16 @@ cocoa_set_activation_policy(bool hide_from_tasks) {
     [NSApp setActivationPolicy:(hide_from_tasks ? NSApplicationActivationPolicyAccessory : NSApplicationActivationPolicyRegular)];
 }
 
+void
+cocoa_set_titlebar_appearance(void *w, unsigned int theme)
+{
+    if (!theme) return;
+    @autoreleasepool {
+        NSWindow *window = (NSWindow*)w;
+        [window setAppearance:[NSAppearance appearanceNamed:((theme == 2) ? NSAppearanceNameVibrantDark : NSAppearanceNameVibrantLight)]];
+    } // autoreleasepool
+}
+
 void
 cocoa_set_titlebar_color(void *w, color_type titlebar_color)
 {
@@ -586,13 +838,43 @@ cocoa_set_titlebar_color(void *w, color_type titlebar_color)
     } // autoreleasepool
 }
 
+static PyObject*
+cocoa_set_url_handler(PyObject UNUSED *self, PyObject *args) {
+    @autoreleasepool {
+
+    const char *url_scheme = NULL, *bundle_id = NULL;
+    if (!PyArg_ParseTuple(args, "s|z", &url_scheme, &bundle_id)) return NULL;
+    if (!url_scheme || url_scheme[0] == '\0') {
+        PyErr_SetString(PyExc_TypeError, "Empty url scheme");
+        return NULL;
+    }
+
+    NSString *scheme = [NSString stringWithUTF8String:url_scheme];
+    NSString *identifier = @"";
+    if (!bundle_id) {
+        identifier = [[NSBundle mainBundle] bundleIdentifier];
+        if (!identifier || identifier.length == 0) identifier = @"net.kovidgoyal.kitty";
+    } else if (bundle_id[0] != '\0') {
+        identifier = [NSString stringWithUTF8String:bundle_id];
+    }
+    // This API has been marked as deprecated. It will need to be replaced when a new approach is available.
+    OSStatus err = LSSetDefaultHandlerForURLScheme((CFStringRef)scheme, (CFStringRef)identifier);
+    if (err == noErr) Py_RETURN_NONE;
+    PyErr_Format(PyExc_OSError, "Failed to set default handler with error code: %d", err);
+    return NULL;
+    } // autoreleasepool
+}
+
+static NSSound *beep_sound = nil;
+
 static void
 cleanup() {
     @autoreleasepool {
 
     if (dockMenu) [dockMenu release];
     dockMenu = nil;
-    Py_CLEAR(notification_activated_callback);
+    if (beep_sound) [beep_sound release];
+    beep_sound = nil;
 
 #ifndef KITTY_USE_DEPRECATED_MACOS_NOTIFICATION_API
     drain_pending_notifications(NO);
@@ -616,24 +898,37 @@ cocoa_hide_window_title(void *w)
 }
 
 void
-cocoa_system_beep(void) {
-    NSBeep();
+cocoa_system_beep(const char *path) {
+    if (!path) { NSBeep(); return; }
+    static const char *beep_path = NULL;
+    if (beep_path != path) {
+        if (beep_sound) [beep_sound release];
+        beep_sound = [[NSSound alloc] initWithContentsOfFile:@(path) byReference:YES];
+    }
+    if (beep_sound) [beep_sound play];
+    else NSBeep();
 }
 
 static PyMethodDef module_methods[] = {
     {"cocoa_get_lang", (PyCFunction)cocoa_get_lang, METH_NOARGS, ""},
-    {"cocoa_set_new_window_trigger", (PyCFunction)cocoa_set_new_window_trigger, METH_VARARGS, ""},
+    {"cocoa_set_global_shortcut", (PyCFunction)cocoa_set_global_shortcut, METH_VARARGS, ""},
     {"cocoa_send_notification", (PyCFunction)cocoa_send_notification, METH_VARARGS, ""},
     {"cocoa_set_notification_activated_callback", (PyCFunction)set_notification_activated_callback, METH_O, ""},
+    {"cocoa_set_url_handler", (PyCFunction)cocoa_set_url_handler, METH_VARARGS, ""},
     {NULL, NULL, 0, NULL}        /* Sentinel */
 };
 
+static void
+uncaughtExceptionHandler(NSException *exception) {
+    log_error("Unhandled exception in Cocoa: %s", [[exception description] UTF8String]);
+    log_error("Stack trace:\n%s", [[exception.callStackSymbols description] UTF8String]);
+}
+
 bool
 init_cocoa(PyObject *module) {
+    memset(&global_shortcuts, 0, sizeof(global_shortcuts));
     if (PyModule_AddFunctions(module, module_methods) != 0) return false;
-    if (Py_AtExit(cleanup) != 0) {
-        PyErr_SetString(PyExc_RuntimeError, "Failed to register the cocoa_window at exit handler");
-        return false;
-    }
+    register_at_exit_cleanup_func(COCOA_CLEANUP_FUNC, cleanup);
+    NSSetUncaughtExceptionHandler(&uncaughtExceptionHandler);
     return true;
 }
