@@ -11,7 +11,10 @@
 #include <unistd.h>
 #undef _DARWIN_C_SOURCE
 #endif
+
 #include "data-types.h"
+#include "cleanup.h"
+#include "safe-wrappers.h"
 #include "control-codes.h"
 #include "wcwidth-std.h"
 #include "wcswidth.h"
@@ -21,13 +24,17 @@
 #include <signal.h>
 #include <fcntl.h>
 #include <stdio.h>
+#include <locale.h>
+
 #ifdef WITH_PROFILER
 #include <gperftools/profiler.h>
 #endif
+
 #include "monotonic.h"
 
 #ifdef __APPLE__
 #include <libproc.h>
+#include <xlocale.h>
 
 static PyObject*
 user_cache_dir() {
@@ -40,18 +47,17 @@ static PyObject*
 process_group_map() {
     int num_of_processes = proc_listallpids(NULL, 0);
     size_t bufsize = sizeof(pid_t) * (num_of_processes + 1024);
-    pid_t *buf = malloc(bufsize);
+    FREE_AFTER_FUNCTION pid_t *buf = malloc(bufsize);
     if (!buf) return PyErr_NoMemory();
     num_of_processes = proc_listallpids(buf, (int)bufsize);
     PyObject *ans = PyTuple_New(num_of_processes);
-    if (ans == NULL) { free(buf); return PyErr_NoMemory(); }
+    if (ans == NULL) { return PyErr_NoMemory(); }
     for (int i = 0; i < num_of_processes; i++) {
         long pid = buf[i], pgid = getpgid(buf[i]);
         PyObject *t = Py_BuildValue("ll", pid, pgid);
-        if (t == NULL) { free(buf); Py_DECREF(ans); return NULL; }
+        if (t == NULL) { Py_DECREF(ans); return NULL; }
         PyTuple_SET_ITEM(ans, i, t);
     }
-    free(buf);
     return ans;
 }
 #endif
@@ -90,7 +96,7 @@ stop_profiler(PyObject UNUSED *self, PyObject *args UNUSED) {
 }
 #endif
 
-static inline bool
+static bool
 put_tty_in_raw_mode(int fd, const struct termios* termios_p, bool read_with_timeout, int optional_actions) {
     struct termios raw_termios = *termios_p;
     cfmakeraw(&raw_termios);
@@ -110,7 +116,7 @@ open_tty(PyObject *self UNUSED, PyObject *args) {
     int flags = O_RDWR | O_CLOEXEC | O_NOCTTY;
     if (!read_with_timeout) flags |= O_NONBLOCK;
     static char ctty[L_ctermid+1];
-    int fd = open(ctermid(ctty), flags);
+    int fd = safe_open(ctermid(ctty), flags, 0);
     if (fd == -1) { PyErr_Format(PyExc_OSError, "Failed to open controlling terminal: %s (identified with ctermid()) with error: %s", ctty, strerror(errno)); return NULL; }
     struct termios *termios_p = calloc(1, sizeof(struct termios));
     if (!termios_p) return PyErr_NoMemory();
@@ -155,6 +161,15 @@ wcwidth_wrap(PyObject UNUSED *self, PyObject *chr) {
     return PyLong_FromLong(wcwidth_std(PyLong_AsLong(chr)));
 }
 
+static PyObject*
+locale_is_valid(PyObject *self UNUSED, PyObject *args) {
+    char *name;
+    if (!PyArg_ParseTuple(args, "s", &name)) return NULL;
+    locale_t test_locale = newlocale(LC_ALL_MASK, name, NULL);
+    if (!test_locale) { Py_RETURN_FALSE; }
+    freelocale(test_locale);
+    Py_RETURN_TRUE;
+}
 
 static PyMethodDef module_methods[] = {
     {"wcwidth", (PyCFunction)wcwidth_wrap, METH_O, ""},
@@ -168,6 +183,7 @@ static PyMethodDef module_methods[] = {
     {"parse_bytes", (PyCFunction)parse_bytes, METH_VARARGS, ""},
     {"parse_bytes_dump", (PyCFunction)parse_bytes_dump, METH_VARARGS, ""},
     {"redirect_std_streams", (PyCFunction)redirect_std_streams, METH_VARARGS, ""},
+    {"locale_is_valid", (PyCFunction)locale_is_valid, METH_VARARGS, ""},
 #ifdef __APPLE__
     METHODB(user_cache_dir, METH_NOARGS),
     METHODB(process_group_map, METH_NOARGS),
@@ -192,6 +208,7 @@ static struct PyModuleDef module = {
 extern int init_LineBuf(PyObject *);
 extern int init_HistoryBuf(PyObject *);
 extern int init_Cursor(PyObject *);
+extern int init_DiskCache(PyObject *);
 extern bool init_child_monitor(PyObject *);
 extern int init_Line(PyObject *);
 extern int init_ColorProfile(PyObject *);
@@ -209,21 +226,43 @@ extern bool init_mouse(PyObject *module);
 extern bool init_kittens(PyObject *module);
 extern bool init_logging(PyObject *module);
 extern bool init_png_reader(PyObject *module);
+extern bool init_utmp(PyObject *module);
 #ifdef __APPLE__
 extern int init_CoreText(PyObject *);
 extern bool init_cocoa(PyObject *module);
 extern bool init_macos_process_info(PyObject *module);
 #else
 extern bool init_freetype_library(PyObject*);
+extern bool init_freetype_render_ui_text(PyObject*);
 #endif
 
+static unsigned
+shift_to_first_set_bit(CellAttrs x) {
+    unsigned num_of_bits = 8 * sizeof(x.val);
+    unsigned ans = 0;
+    while (num_of_bits--) {
+        if (x.val & 1) return ans;
+        x.val >>= 1;
+        ans++;
+    }
+    return ans;
+}
+
 
 EXPORTED PyMODINIT_FUNC
 PyInit_fast_data_types(void) {
     PyObject *m;
+    if (sizeof(CellAttrs) != 2u) {
+        PyErr_SetString(PyExc_RuntimeError, "Size of CellAttrs is not 2 on this platform");
+        return NULL;
+    }
 
     m = PyModule_Create(&module);
     if (m == NULL) return NULL;
+    if (Py_AtExit(run_at_exit_cleanup_functions) != 0) {
+        PyErr_SetString(PyExc_RuntimeError, "Failed to register the atexit cleanup handler");
+        return NULL;
+    }
     init_monotonic();
 
     if (!init_logging(m)) return NULL;
@@ -231,6 +270,7 @@ PyInit_fast_data_types(void) {
     if (!init_HistoryBuf(m)) return NULL;
     if (!init_Line(m)) return NULL;
     if (!init_Cursor(m)) return NULL;
+    if (!init_DiskCache(m)) return NULL;
     if (!init_child_monitor(m)) return NULL;
     if (!init_ColorProfile(m)) return NULL;
     if (!init_Screen(m)) return NULL;
@@ -251,17 +291,19 @@ PyInit_fast_data_types(void) {
     if (!init_freetype_library(m)) return NULL;
     if (!init_fontconfig_library(m)) return NULL;
     if (!init_desktop(m)) return NULL;
+    if (!init_freetype_render_ui_text(m)) return NULL;
 #endif
     if (!init_fonts(m)) return NULL;
+    if (!init_utmp(m)) return NULL;
 
-    PyModule_AddIntConstant(m, "BOLD", BOLD_SHIFT);
-    PyModule_AddIntConstant(m, "ITALIC", ITALIC_SHIFT);
-    PyModule_AddIntConstant(m, "REVERSE", REVERSE_SHIFT);
-    PyModule_AddIntConstant(m, "STRIKETHROUGH", STRIKE_SHIFT);
-    PyModule_AddIntConstant(m, "DIM", DIM_SHIFT);
-    PyModule_AddIntConstant(m, "DECORATION", DECORATION_SHIFT);
-    PyModule_AddIntConstant(m, "MARK", MARK_SHIFT);
+    CellAttrs a;
+#define s(name, attr) { a.val = 0; a.attr = 1; PyModule_AddIntConstant(m, #name, shift_to_first_set_bit(a)); }
+    s(BOLD, bold); s(ITALIC, italic); s(REVERSE, reverse); s(MARK, mark);
+    s(STRIKETHROUGH, strike); s(DIM, dim); s(DECORATION, decoration);
+#undef s
     PyModule_AddIntConstant(m, "MARK_MASK", MARK_MASK);
+    PyModule_AddIntConstant(m, "DECORATION_MASK", DECORATION_MASK);
+    PyModule_AddIntConstant(m, "NUM_UNDERLINE_STYLES", NUM_UNDERLINE_STYLES);
     PyModule_AddStringMacro(m, ERROR_PREFIX);
 #ifdef KITTY_VCS_REV
     PyModule_AddStringMacro(m, KITTY_VCS_REV);
@@ -269,6 +311,7 @@ PyInit_fast_data_types(void) {
     PyModule_AddIntMacro(m, CURSOR_BLOCK);
     PyModule_AddIntMacro(m, CURSOR_BEAM);
     PyModule_AddIntMacro(m, CURSOR_UNDERLINE);
+    PyModule_AddIntMacro(m, NO_CURSOR_SHAPE);
     PyModule_AddIntMacro(m, DECAWM);
     PyModule_AddIntMacro(m, DECCOLM);
     PyModule_AddIntMacro(m, DECOM);
@@ -277,6 +320,7 @@ PyInit_fast_data_types(void) {
     PyModule_AddIntMacro(m, DCS);
     PyModule_AddIntMacro(m, APC);
     PyModule_AddIntMacro(m, OSC);
+    PyModule_AddIntMacro(m, FILE_TRANSFER_CODE);
 
     return m;
 }
