@@ -6,21 +6,17 @@
  */
 
 #include "fonts.h"
+#include "cleanup.h"
 #include "state.h"
 #include <math.h>
 #include <structmember.h>
 #include <ft2build.h>
 #include <hb-ft.h>
 
-#if HB_VERSION_MAJOR > 1 || (HB_VERSION_MAJOR == 1 && (HB_VERSION_MINOR > 6 || (HB_VERSION_MINOR == 6 && HB_VERSION_MICRO >= 3)))
-#define HARFBUZZ_HAS_CHANGE_FONT
-#endif
-
 #if FREETYPE_MAJOR == 2 && FREETYPE_MINOR < 7
 #define FT_Bitmap_Init FT_Bitmap_New
 #endif
 
-#include FT_FREETYPE_H
 #include FT_BITMAP_H
 #include FT_TRUETYPE_TABLES_H
 typedef struct {
@@ -36,6 +32,7 @@ typedef struct {
     FT_UInt xdpi, ydpi;
     PyObject *path;
     hb_font_t *harfbuzz_font;
+    hb_codepoint_t space_glyph_id;
     void *extra_data;
     free_extra_data_func free_extra_data;
     float apple_leading;
@@ -76,37 +73,44 @@ set_freetype_error(const char* prefix, int err_code) {
 
 static FT_Library  library;
 
-static inline int
+FT_Library
+freetype_library(void) { return library; }
+
+static int
 font_units_to_pixels_y(Face *self, int x) {
     return (int)ceil((double)FT_MulFix(x, self->face->size->metrics.y_scale) / 64.0);
 }
 
-static inline int
+static int
 font_units_to_pixels_x(Face *self, int x) {
     return (int)ceil((double)FT_MulFix(x, self->face->size->metrics.x_scale) / 64.0);
 }
 
 
-static inline int
+static int
 get_load_flags(int hinting, int hintstyle, int base) {
     int flags = base;
     if (hinting) {
         if (hintstyle >= 3) flags |= FT_LOAD_TARGET_NORMAL;
-        else if (0 < hintstyle  && hintstyle < 3) flags |= FT_LOAD_TARGET_LIGHT;
+        else if (0 < hintstyle) flags |= FT_LOAD_TARGET_LIGHT;
     } else flags |= FT_LOAD_NO_HINTING;
     return flags;
 }
 
 
-static inline bool
+static bool
 load_glyph(Face *self, int glyph_index, int load_type) {
     int flags = get_load_flags(self->hinting, self->hintstyle, load_type);
     int error = FT_Load_Glyph(self->face, glyph_index, flags);
-    if (error) { set_freetype_error("Failed to load glyph, with error:", error); return false; }
+    if (error) {
+        char buf[256];
+        snprintf(buf, sizeof(buf) - 1, "Failed to load glyph_index=%d load_type=%d, with error:", glyph_index, load_type);
+        set_freetype_error(buf, error); return false;
+    }
     return true;
 }
 
-static inline unsigned int
+static unsigned int
 get_height_for_char(Face *self, char ch) {
     unsigned int ans = 0;
     int glyph_index = FT_Get_Char_Index(self->face, ch);
@@ -121,7 +125,7 @@ get_height_for_char(Face *self, char ch) {
     return ans;
 }
 
-static inline unsigned int
+static unsigned int
 calc_cell_height(Face *self, bool for_metrics) {
     unsigned int ans = font_units_to_pixels_y(self, self->height);
     if (for_metrics) {
@@ -135,27 +139,12 @@ calc_cell_height(Face *self, bool for_metrics) {
     return ans;
 }
 
-static inline bool
+static bool
 set_font_size(Face *self, FT_F26Dot6 char_width, FT_F26Dot6 char_height, FT_UInt xdpi, FT_UInt ydpi, unsigned int desired_height, unsigned int cell_height) {
     int error = FT_Set_Char_Size(self->face, 0, char_height, xdpi, ydpi);
     if (!error) {
-        unsigned int ch = calc_cell_height(self, false);
-        if (desired_height && ch != desired_height) {
-            FT_F26Dot6 h = (FT_F26Dot6)floor((double)char_height * (double)desired_height / (double) ch);
-            return set_font_size(self, 0, h, xdpi, ydpi, 0, cell_height);
-        }
         self->char_width = char_width; self->char_height = char_height; self->xdpi = xdpi; self->ydpi = ydpi;
-        if (self->harfbuzz_font != NULL) {
-#ifdef HARFBUZZ_HAS_CHANGE_FONT
-            hb_ft_font_changed(self->harfbuzz_font);
-#else
-            hb_font_set_scale(
-                self->harfbuzz_font,
-                (int) (((uint64_t) self->face->size->metrics.x_scale * (uint64_t) self->face->units_per_EM + (1u<<15)) >> 16),
-                (int) (((uint64_t) self->face->size->metrics.y_scale * (uint64_t) self->face->units_per_EM + (1u<<15)) >> 16)
-            );
-#endif
-        }
+        if (self->harfbuzz_font != NULL) hb_ft_font_changed(self->harfbuzz_font);
     } else {
         if (!self->is_scalable && self->face->num_fixed_sizes > 0) {
             int32_t min_diff = INT32_MAX;
@@ -195,7 +184,7 @@ set_size_for_face(PyObject *s, unsigned int desired_height, bool force, FONTS_DA
     return set_font_size(self, w, w, xdpi, ydpi, desired_height, fg->cell_height);
 }
 
-static inline bool
+static bool
 init_ft_face(Face *self, PyObject *path, int hinting, int hintstyle, FONTS_DATA_HANDLE fg) {
 #define CPY(n) self->n = self->face->n;
     CPY(units_per_EM); CPY(ascender); CPY(descender); CPY(height); CPY(max_advance_width); CPY(max_advance_height); CPY(underline_position); CPY(underline_thickness);
@@ -217,6 +206,7 @@ init_ft_face(Face *self, PyObject *path, int hinting, int hintstyle, FONTS_DATA_
     self->path = path;
     Py_INCREF(self->path);
     self->index = self->face->face_index & 0xFFFF;
+    self->space_glyph_id = glyph_id_for_codepoint((PyObject*)self, ' ');
     return true;
 }
 
@@ -246,6 +236,15 @@ face_from_descriptor(PyObject *descriptor, FONTS_DATA_HANDLE fg) {
     return (PyObject*)self;
 }
 
+FT_Face
+native_face_from_path(const char *path, int index) {
+    int error;
+    FT_Face ans;
+    error = FT_New_Face(library, path, index, &ans);
+    if (error) { set_freetype_error("Failed to load face, with error:", error); return NULL; }
+    return ans;
+}
+
 PyObject*
 face_from_path(const char *path, int index, FONTS_DATA_HANDLE fg) {
     Face *ans = (Face*)Face_Type.tp_alloc(&Face_Type, 0);
@@ -286,7 +285,7 @@ postscript_name_for_face(const PyObject *face_) {
     return ps_name ? ps_name : "";
 }
 
-static inline unsigned int
+static unsigned int
 calc_cell_width(Face *self) {
     unsigned int ans = 0;
     for (char_type i = 32; i < 128; i++) {
@@ -298,11 +297,22 @@ calc_cell_width(Face *self) {
     return ans;
 }
 
+
+static unsigned int
+adjust_ypos(unsigned int pos, unsigned int cell_height, int adjustment) {
+    if (adjustment >= 0) adjustment = MIN(adjustment, (int)pos - 1);
+    else adjustment = MAX(adjustment, (int)pos - (int)cell_height + 1);
+    return pos - adjustment;
+}
+
 void
 cell_metrics(PyObject *s, unsigned int* cell_width, unsigned int* cell_height, unsigned int* baseline, unsigned int* underline_position, unsigned int* underline_thickness, unsigned int* strikethrough_position, unsigned int* strikethrough_thickness) {
     Face *self = (Face*)s;
     *cell_width = calc_cell_width(self);
     *cell_height = calc_cell_height(self, true);
+    int baseline_offset = 0;
+    if (OPT(adjust_baseline_px) != 0) baseline_offset = OPT(adjust_baseline_px);
+    else if (OPT(adjust_baseline_frac) != 0) baseline_offset = (int)(*cell_height * OPT(adjust_baseline_frac));
     *baseline = font_units_to_pixels_y(self, self->ascender);
     *underline_position = MIN(*cell_height - 1, (unsigned int)font_units_to_pixels_y(self, MAX(0, self->ascender - self->underline_position)));
     *underline_thickness = MAX(1, font_units_to_pixels_y(self, self->underline_thickness));
@@ -317,6 +327,11 @@ cell_metrics(PyObject *s, unsigned int* cell_width, unsigned int* cell_height, u
     } else {
       *strikethrough_thickness = *underline_thickness;
     }
+    if (baseline_offset) {
+        *baseline = adjust_ypos(*baseline, *cell_height, baseline_offset);
+        *underline_position = adjust_ypos(*underline_position, *cell_height, baseline_offset);
+        *strikethrough_position = adjust_ypos(*strikethrough_position, *cell_height, baseline_offset);
+    }
 }
 
 unsigned int
@@ -338,7 +353,7 @@ is_glyph_empty(PyObject *s, glyph_index g) {
 int
 get_glyph_width(PyObject *s, glyph_index g) {
     Face *self = (Face*)s;
-    if (!load_glyph(self, g, FT_LOAD_DEFAULT)) { PyErr_Print(); return false; }
+    if (!load_glyph(self, g, FT_LOAD_DEFAULT)) { PyErr_Print(); return 0; }
 #define M self->face->glyph->metrics
 #define B self->face->glyph->bitmap
     /* printf("glyph: %u bitmap.width: %d bitmap.rows: %d horiAdvance: %ld horiBearingX: %ld horiBearingY: %ld vertBearingX: %ld vertBearingY: %ld vertAdvance: %ld width: %ld height: %ld\n", */
@@ -362,7 +377,7 @@ typedef struct {
     int bitmap_left, bitmap_top;
 } ProcessedBitmap;
 
-static inline void
+static void
 free_processed_bitmap(ProcessedBitmap *bm) {
     if (bm->needs_free) {
         bm->needs_free = false;
@@ -370,7 +385,7 @@ free_processed_bitmap(ProcessedBitmap *bm) {
     }
 }
 
-static inline void
+static void
 trim_borders(ProcessedBitmap *ans, size_t extra) {
     bool column_has_text = false;
 
@@ -387,7 +402,7 @@ trim_borders(ProcessedBitmap *ans, size_t extra) {
     ans->width -= extra;
 }
 
-static inline void
+static void
 populate_processed_bitmap(FT_GlyphSlotRec *slot, FT_Bitmap *bitmap, ProcessedBitmap *ans, bool copy_buf) {
     ans->stride = bitmap->pitch < 0 ? -bitmap->pitch : bitmap->pitch;
     ans->rows = bitmap->rows;
@@ -402,7 +417,23 @@ populate_processed_bitmap(FT_GlyphSlotRec *slot, FT_Bitmap *bitmap, ProcessedBit
     ans->bitmap_top = slot->bitmap_top; ans->bitmap_left = slot->bitmap_left;
 }
 
-static inline bool
+bool
+freetype_convert_mono_bitmap(FT_Bitmap *src, FT_Bitmap *dest) {
+    FT_Bitmap_Init(dest);
+    // This also sets pixel_mode to FT_PIXEL_MODE_GRAY so we don't have to
+    int error = FT_Bitmap_Convert(library, src, dest, 1);
+    if (error) { set_freetype_error("Failed to convert bitmap, with error:", error); return false; }
+    // Normalize gray levels to the range [0..255]
+    dest->num_grays = 256;
+    unsigned int stride = dest->pitch < 0 ? -dest->pitch : dest->pitch;
+    for (unsigned i = 0; i < (unsigned)dest->rows; ++i) {
+        // We only have 2 levels
+        for (unsigned j = 0; j < (unsigned)dest->width; ++j) dest->buffer[i * stride + j] *= 255;
+    }
+    return true;
+}
+
+static bool
 render_bitmap(Face *self, int glyph_id, ProcessedBitmap *ans, unsigned int cell_width, unsigned int cell_height, unsigned int num_cells, bool bold, bool italic, bool rescale, FONTS_DATA_HANDLE fg) {
     if (!load_glyph(self, glyph_id, FT_LOAD_RENDER)) return false;
     unsigned int max_width = cell_width * num_cells;
@@ -410,19 +441,7 @@ render_bitmap(Face *self, int glyph_id, ProcessedBitmap *ans, unsigned int cell_
     // Embedded bitmap glyph?
     if (self->face->glyph->bitmap.pixel_mode == FT_PIXEL_MODE_MONO) {
         FT_Bitmap bitmap;
-        FT_Bitmap_Init(&bitmap);
-
-        // This also sets pixel_mode to FT_PIXEL_MODE_GRAY so we don't have to
-        int error = FT_Bitmap_Convert(library, &self->face->glyph->bitmap, &bitmap, 1);
-        if (error) { set_freetype_error("Failed to convert bitmap, with error:", error); return false; }
-
-        // Normalize gray levels to the range [0..255]
-        bitmap.num_grays = 256;
-        unsigned int stride = bitmap.pitch < 0 ? -bitmap.pitch : bitmap.pitch;
-        for (unsigned i = 0; i < (unsigned)bitmap.rows; ++i) {
-            // We only have 2 levels
-            for (unsigned j = 0; j < (unsigned)bitmap.width; ++j) bitmap.buffer[i * stride + j] *= 255;
-        }
+        freetype_convert_mono_bitmap(&self->face->glyph->bitmap, &bitmap);
         populate_processed_bitmap(self->face->glyph, &bitmap, ans, true);
         FT_Bitmap_Done(library, &bitmap);
     } else {
@@ -450,39 +469,41 @@ render_bitmap(Face *self, int glyph_id, ProcessedBitmap *ans, unsigned int cell_
     return true;
 }
 
-static void
-downsample_bitmap(ProcessedBitmap *bm, unsigned int width, unsigned int cell_height) {
+int
+downsample_32bit_image(uint8_t *src, unsigned src_width, unsigned src_height, unsigned src_stride, uint8_t *dest, unsigned dest_width, unsigned dest_height) {
     // Downsample using a simple area averaging algorithm. Could probably do
     // better with bi-cubic or lanczos, but at these small sizes I don't think
     // it matters
-    float ratio = MAX((float)bm->width / width, (float)bm->rows / cell_height);
+    float ratio = MAX((float)src_width / dest_width, (float)src_height / dest_height);
     int factor = (int)ceilf(ratio);
-    uint8_t *dest = calloc(4, (size_t)width * cell_height);
-    if (dest == NULL) fatal("Out of memory");
     uint8_t *d = dest;
-
-    for (unsigned int i = 0, sr = 0; i < cell_height; i++, sr += factor) {
-        for (unsigned int j = 0, sc = 0; j < width; j++, sc += factor, d += 4) {
-
+    for (unsigned int i = 0, sr = 0; i < dest_height; i++, sr += factor) {
+        for (unsigned int j = 0, sc = 0; j < dest_width; j++, sc += factor, d += 4) {
             // calculate area average
             unsigned int r=0, g=0, b=0, a=0, count=0;
-            for (unsigned int y=sr; y < MIN(sr + factor, bm->rows); y++) {
-                uint8_t *p = bm->buf + (y * bm->stride) + sc * 4;
-                for (unsigned int x=sc; x < MIN(sc + factor, bm->width); x++, count++) {
+            for (unsigned int y=sr; y < MIN(sr + factor, src_height); y++) {
+                uint8_t *p = src + (y * src_stride) + sc * 4;
+                for (unsigned int x=sc; x < MIN(sc + factor, src_width); x++, count++) {
                     b += *(p++); g += *(p++); r += *(p++); a += *(p++);
                 }
             }
             if (count) {
                 d[0] = b / count; d[1] = g / count; d[2] = r / count; d[3] = a / count;
             }
-
         }
     }
-    bm->buf = dest; bm->needs_free = true; bm->stride = 4 * width; bm->width = width; bm->rows = cell_height;
-    bm->factor = factor;
+    return factor;
 }
 
-static inline void
+static void
+downsample_bitmap(ProcessedBitmap *bm, unsigned int width, unsigned int cell_height) {
+    uint8_t *dest = calloc(4, (size_t)width * cell_height);
+    if (dest == NULL) fatal("Out of memory");
+    bm->factor = downsample_32bit_image(bm->buf, bm->width, bm->rows, bm->stride, dest, width, cell_height);
+    bm->buf = dest; bm->needs_free = true; bm->stride = 4 * width; bm->width = width; bm->rows = cell_height;
+}
+
+static void
 detect_right_edge(ProcessedBitmap *ans) {
     ans->right_edge = 0;
     for (ssize_t x = ans->width - 1; !ans->right_edge && x > -1; x--) {
@@ -493,23 +514,23 @@ detect_right_edge(ProcessedBitmap *ans) {
     }
 }
 
-static inline bool
-render_color_bitmap(Face *self, int glyph_id, ProcessedBitmap *ans, unsigned int cell_width, unsigned int cell_height, unsigned int num_cells, unsigned int baseline) {
-    (void)baseline;
+static bool
+render_color_bitmap(Face *self, int glyph_id, ProcessedBitmap *ans, unsigned int cell_width, unsigned int cell_height, unsigned int num_cells, unsigned int baseline UNUSED) {
     unsigned short best = 0, diff = USHRT_MAX;
-    const short limit = self->face->num_fixed_sizes;
-    for (short i = 0; i < limit; i++) {
-        unsigned short w = self->face->available_sizes[i].width;
-        unsigned short d = w > (unsigned short)cell_width ? w - (unsigned short)cell_width : (unsigned short)cell_width - w;
-        if (d < diff) {
-            diff = d;
-            best = i;
+    if (self->face->num_fixed_sizes > 0) {
+        const short limit = self->face->num_fixed_sizes;
+        for (short i = 0; i < limit; i++) {
+            unsigned short w = self->face->available_sizes[i].width;
+            unsigned short d = w > (unsigned short)cell_width ? w - (unsigned short)cell_width : (unsigned short)cell_width - w;
+            if (d < diff) {
+                diff = d;
+                best = i;
+            }
         }
-    }
-    FT_Error error = FT_Select_Size(self->face, best);
-    if (error) { set_freetype_error("Failed to set char size for non-scalable font, with error:", error); return false; }
-    if (!load_glyph(self, glyph_id, FT_LOAD_COLOR)) return false;
-    FT_Set_Char_Size(self->face, 0, self->char_height, self->xdpi, self->ydpi);
+        FT_Error error = FT_Select_Size(self->face, best);
+        if (error) { set_freetype_error("Failed to set char size for non-scalable font, with error:", error); return false; }
+    } else FT_Set_Char_Size(self->face, 0, self->char_height, self->xdpi, self->ydpi);
+    if (!load_glyph(self, glyph_id, FT_LOAD_COLOR | FT_LOAD_RENDER)) return false;
     FT_Bitmap *bitmap = &self->face->glyph->bitmap;
     if (bitmap->pixel_mode != FT_PIXEL_MODE_BGRA) return false;
     ans->buf = bitmap->buffer;
@@ -525,7 +546,7 @@ render_color_bitmap(Face *self, int glyph_id, ProcessedBitmap *ans, unsigned int
 }
 
 
-static inline void
+static void
 copy_color_bitmap(uint8_t *src, pixel* dest, Region *src_rect, Region *dest_rect, size_t src_stride, size_t dest_stride) {
     for (size_t sr = src_rect->top, dr = dest_rect->top; sr < src_rect->bottom && dr < dest_rect->bottom; sr++, dr++) {
         pixel *d = dest + dest_stride * dr;
@@ -541,21 +562,23 @@ copy_color_bitmap(uint8_t *src, pixel* dest, Region *src_rect, Region *dest_rect
     }
 }
 
-static inline void
-place_bitmap_in_canvas(pixel *cell, ProcessedBitmap *bm, size_t cell_width, size_t cell_height, float x_offset, float y_offset, size_t baseline) {
+static const bool debug_placement = false;
+
+static void
+place_bitmap_in_canvas(pixel *cell, ProcessedBitmap *bm, size_t cell_width, size_t cell_height, float x_offset, float y_offset, size_t baseline, unsigned int glyph_num) {
     // We want the glyph to be positioned inside the cell based on the bearingX
     // and bearingY values, making sure that it does not overflow the cell.
 
     Region src = { .left = bm->start_x, .bottom = bm->rows, .right = bm->width + bm->start_x }, dest = { .bottom = cell_height, .right = cell_width };
 
     // Calculate column bounds
-    int32_t xoff = (ssize_t)(x_offset + bm->bitmap_left);
-    uint32_t extra;
+    int32_t xoff = (int32_t)(x_offset + bm->bitmap_left);
+    if (debug_placement) printf(" bitmap_left: %d xoff: %d", bm->bitmap_left, xoff);
     if (xoff < 0) src.left += -xoff;
     else dest.left = xoff;
-    // Move the dest start column back if the width overflows because of it
-    if (dest.left > 0 && dest.left + bm->width > cell_width) {
-        extra = dest.left + bm->width - cell_width;
+    // Move the dest start column back if the width overflows because of it, but only if we are not in a very long/infinite ligature
+    if (glyph_num < 4 && dest.left > 0 && dest.left + bm->width > cell_width) {
+        uint32_t extra = dest.left + bm->width - cell_width;
         dest.left = extra > dest.left ? 0 : dest.left - extra;
     }
 
@@ -567,7 +590,7 @@ place_bitmap_in_canvas(pixel *cell, ProcessedBitmap *bm, size_t cell_width, size
         dest.top = baseline - yoff;
     }
 
-    /* printf("x_offset: %d bearing_x: %f y_offset: %d bearing_y: %f src_start_row: %u src_start_column: %u dest_start_row: %u dest_start_column: %u bm_width: %lu bitmap_rows: %lu\n", xoff, bearing_x, yoff, bearing_y, src.top, src.left, dest.top, dest.left, bm->width, bm->rows); */
+    /* printf("x_offset: %d y_offset: %d src_start_row: %u src_start_column: %u dest_start_row: %u dest_start_column: %u bm_width: %lu bitmap_rows: %lu\n", xoff, yoff, src.top, src.left, dest.top, dest.left, bm->width, bm->rows); */
 
     if (bm->pixel_mode == FT_PIXEL_MODE_BGRA) {
         copy_color_bitmap(bm->buf, cell, &src, &dest, bm->stride, cell_width);
@@ -585,27 +608,37 @@ render_glyphs_in_cells(PyObject *f, bool bold, bool italic, hb_glyph_info_t *inf
     unsigned int canvas_width = cell_width * num_cells;
     for (unsigned int i = 0; i < num_glyphs; i++) {
         bm = EMPTY_PBM;
-        if (*was_colored) {
-            if (!render_color_bitmap(self, info[i].codepoint, &bm, cell_width, cell_height, num_cells, baseline)) {
-                if (PyErr_Occurred()) PyErr_Print();
-                *was_colored = false;
+        // dont load the space glyph since loading it fails for some fonts/sizes and it is anyway to be rendered as a blank
+        if (info[i].codepoint != self->space_glyph_id) {
+            if (*was_colored) {
+                if (!render_color_bitmap(self, info[i].codepoint, &bm, cell_width, cell_height, num_cells, baseline)) {
+                    if (PyErr_Occurred()) PyErr_Print();
+                    if (!render_bitmap(self, info[i].codepoint, &bm, cell_width, cell_height, num_cells, bold, italic, true, fg)) {
+                        free_processed_bitmap(&bm);
+                        return false;
+                    }
+                    *was_colored = false;
+                }
+            } else {
                 if (!render_bitmap(self, info[i].codepoint, &bm, cell_width, cell_height, num_cells, bold, italic, true, fg)) {
                     free_processed_bitmap(&bm);
                     return false;
                 }
             }
-        } else {
-            if (!render_bitmap(self, info[i].codepoint, &bm, cell_width, cell_height, num_cells, bold, italic, true, fg)) {
-                free_processed_bitmap(&bm);
-                return false;
-            }
         }
         x_offset = x + (float)positions[i].x_offset / 64.0f;
         y = (float)positions[i].y_offset / 64.0f;
+        if (debug_placement) printf("%d: x=%f canvas: %u", i, x_offset, canvas_width);
         if ((*was_colored || self->face->glyph->metrics.width > 0) && bm.width > 0) {
-            place_bitmap_in_canvas(canvas, &bm, canvas_width, cell_height, x_offset, y, baseline);
+            place_bitmap_in_canvas(canvas, &bm, canvas_width, cell_height, x_offset, y, baseline, i);
         }
-        x += (float)positions[i].x_advance / 64.0f;
+        if (debug_placement) printf(" adv: %f\n", (float)positions[i].x_advance / 64.0f);
+        // the roundf() below is needed for infinite length ligatures, for a test case
+        // use: kitty --config None -o 'font_family Fira Code' -o 'font_size 4.5' sh -c
+        // "echo '|---|--------|-------|-------------|-------------|HH'; read"
+        // if this causes issues with non-infinite ligatures, we could choose this behavior
+        // based on num_glyphs and/or num_cells
+        x += roundf((float)positions[i].x_advance / 64.0f);
         free_processed_bitmap(&bm);
     }
 
@@ -656,7 +689,7 @@ render_simple_text_impl(PyObject *s, const char *text, unsigned int baseline) {
         FT_Bitmap *bitmap = &self->face->glyph->bitmap;
         pbm = EMPTY_PBM;
         populate_processed_bitmap(self->face->glyph, bitmap, &pbm, false);
-        place_bitmap_in_canvas(canvas, &pbm, canvas_width, canvas_height, pen_x, 0, baseline);
+        place_bitmap_in_canvas(canvas, &pbm, canvas_width, canvas_height, pen_x, 0, baseline, n);
         pen_x += self->face->glyph->advance.x >> 6;
     }
     ans.width = pen_x; ans.height = canvas_height;
@@ -728,10 +761,7 @@ init_freetype_library(PyObject *m) {
         set_freetype_error("Failed to initialize FreeType library, with error:", error);
         return false;
     }
-    if (Py_AtExit(free_freetype) != 0) {
-        PyErr_SetString(FreeType_Exception, "Failed to register the freetype library at exit handler");
-        return false;
-    }
+    register_at_exit_cleanup_func(FREETYPE_CLEANUP_FUNC, free_freetype);
     return true;
 }
 
