@@ -14,19 +14,26 @@
 extern PyTypeObject Line_Type;
 #define SEGMENT_SIZE 2048
 
-static inline void
+static void
 add_segment(HistoryBuf *self) {
     self->num_segments += 1;
-    self->segments = PyMem_Realloc(self->segments, sizeof(HistoryBufSegment) * self->num_segments);
+    self->segments = realloc(self->segments, sizeof(HistoryBufSegment) * self->num_segments);
     if (self->segments == NULL) fatal("Out of memory allocating new history buffer segment");
     HistoryBufSegment *s = self->segments + self->num_segments - 1;
-    s->cpu_cells = PyMem_Calloc(self->xnum * SEGMENT_SIZE, sizeof(CPUCell));
-    s->gpu_cells = PyMem_Calloc(self->xnum * SEGMENT_SIZE, sizeof(GPUCell));
-    s->line_attrs = PyMem_Calloc(SEGMENT_SIZE, sizeof(line_attrs_type));
-    if (s->cpu_cells == NULL || s->gpu_cells == NULL || s->line_attrs == NULL) fatal("Out of memory allocating new history buffer segment");
+    const size_t cpu_cells_size = self->xnum * SEGMENT_SIZE * sizeof(CPUCell);
+    const size_t gpu_cells_size = self->xnum * SEGMENT_SIZE * sizeof(GPUCell);
+    s->cpu_cells = calloc(1, cpu_cells_size + gpu_cells_size + SEGMENT_SIZE * sizeof(LineAttrs));
+    if (!s->cpu_cells) fatal("Out of memory allocating new history buffer segment");
+    s->gpu_cells = (GPUCell*)(((uint8_t*)s->cpu_cells) + cpu_cells_size);
+    s->line_attrs = (LineAttrs*)(((uint8_t*)s->gpu_cells) + gpu_cells_size);
 }
 
-static inline index_type
+static void
+free_segment(HistoryBufSegment *s) {
+    free(s->cpu_cells); memset(s, 0, sizeof(HistoryBufSegment));
+}
+
+static index_type
 segment_for(HistoryBuf *self, index_type y) {
     index_type seg_num = y / SEGMENT_SIZE;
     while (UNLIKELY(seg_num >= self->num_segments && SEGMENT_SIZE * self->num_segments < self->ynum)) add_segment(self);
@@ -40,43 +47,46 @@ segment_for(HistoryBuf *self, index_type y) {
     return self->segments[seg_num].which + y * stride; \
 }
 
-static inline CPUCell*
+static CPUCell*
 cpu_lineptr(HistoryBuf *self, index_type y) {
     seg_ptr(cpu_cells, self->xnum);
 }
 
-static inline GPUCell*
+static GPUCell*
 gpu_lineptr(HistoryBuf *self, index_type y) {
     seg_ptr(gpu_cells, self->xnum);
 }
 
 
-static inline line_attrs_type*
+static LineAttrs*
 attrptr(HistoryBuf *self, index_type y) {
     seg_ptr(line_attrs, 1);
 }
 
-static inline PagerHistoryBuf*
+static size_t
+initial_pagerhist_ringbuf_sz(size_t pagerhist_sz) { return MIN(1024u * 1024u, pagerhist_sz); }
+
+static PagerHistoryBuf*
 alloc_pagerhist(size_t pagerhist_sz) {
     PagerHistoryBuf *ph;
     if (!pagerhist_sz) return NULL;
-    ph = PyMem_Calloc(1, sizeof(PagerHistoryBuf));
+    ph = calloc(1, sizeof(PagerHistoryBuf));
     if (!ph) return NULL;
-    size_t sz = MIN(1024u * 1024u, pagerhist_sz);
+    size_t sz = initial_pagerhist_ringbuf_sz(pagerhist_sz);
     ph->ringbuf = ringbuf_new(sz);
-    if (!ph->ringbuf) { PyMem_Free(ph); return NULL; }
+    if (!ph->ringbuf) { free(ph); return NULL; }
     ph->maximum_size = pagerhist_sz;
     return ph;
 }
 
-static inline void
+static void
 free_pagerhist(HistoryBuf *self) {
     if (self->pagerhist && self->pagerhist->ringbuf) ringbuf_free((ringbuf_t*)&self->pagerhist->ringbuf);
-    PyMem_Free(self->pagerhist);
+    free(self->pagerhist);
     self->pagerhist = NULL;
 }
 
-static inline bool
+static bool
 pagerhist_extend(PagerHistoryBuf *ph, size_t minsz) {
     size_t buffer_size = ringbuf_capacity(ph->ringbuf);
     if (buffer_size >= ph->maximum_size) return false;
@@ -90,9 +100,17 @@ pagerhist_extend(PagerHistoryBuf *ph, size_t minsz) {
     return true;
 }
 
-static inline void
+static void
 pagerhist_clear(HistoryBuf *self) {
-    if (self->pagerhist && self->pagerhist->ringbuf) ringbuf_reset(self->pagerhist->ringbuf);
+    if (self->pagerhist && self->pagerhist->ringbuf) {
+        ringbuf_reset(self->pagerhist->ringbuf);
+        size_t rsz = initial_pagerhist_ringbuf_sz(self->pagerhist->maximum_size);
+        void *rbuf = ringbuf_new(rsz);
+        if (rbuf) {
+            ringbuf_free((ringbuf_t*)&self->pagerhist->ringbuf);
+            self->pagerhist->ringbuf = rbuf;
+        }
+    }
 }
 
 static HistoryBuf*
@@ -125,17 +143,13 @@ new(PyTypeObject *type, PyObject *args, PyObject UNUSED *kwds) {
 static void
 dealloc(HistoryBuf* self) {
     Py_CLEAR(self->line);
-    for (size_t i = 0; i < self->num_segments; i++) {
-        PyMem_Free(self->segments[i].cpu_cells);
-        PyMem_Free(self->segments[i].gpu_cells);
-        PyMem_Free(self->segments[i].line_attrs);
-    }
-    PyMem_Free(self->segments);
+    for (size_t i = 0; i < self->num_segments; i++) free_segment(self->segments + i);
+    free(self->segments);
     free_pagerhist(self);
     Py_TYPE(self)->tp_free((PyObject*)self);
 }
 
-static inline index_type
+static index_type
 index_of(HistoryBuf *self, index_type lnum) {
     // The index (buffer position) of the line with line number lnum
     // This is reverse indexing, i.e. lnum = 0 corresponds to the *last* line in the buffer.
@@ -144,13 +158,12 @@ index_of(HistoryBuf *self, index_type lnum) {
     return (self->start_of_data + idx) % self->ynum;
 }
 
-static inline void
+static void
 init_line(HistoryBuf *self, index_type num, Line *l) {
     // Initialize the line l, setting its pointer to the offsets for the line at index (buffer position) num
     l->cpu_cells = cpu_lineptr(self, num);
     l->gpu_cells = gpu_lineptr(self, num);
-    l->continued = *attrptr(self, num) & CONTINUED_MASK;
-    l->has_dirty_text = *attrptr(self, num) & TEXT_DIRTY_MASK ? true : false;
+    l->attrs = *attrptr(self, num);
 }
 
 void
@@ -165,14 +178,12 @@ historybuf_cpu_cells(HistoryBuf *self, index_type lnum) {
 
 void
 historybuf_mark_line_clean(HistoryBuf *self, index_type y) {
-    line_attrs_type *p = attrptr(self, index_of(self, y));
-    *p &= ~TEXT_DIRTY_MASK;
+    attrptr(self, index_of(self, y))->has_dirty_text = false;
 }
 
 void
 historybuf_mark_line_dirty(HistoryBuf *self, index_type y) {
-    line_attrs_type *p = attrptr(self, index_of(self, y));
-    *p |= TEXT_DIRTY_MASK;
+    attrptr(self, index_of(self, y))->has_dirty_text = true;
 }
 
 void
@@ -180,9 +191,11 @@ historybuf_clear(HistoryBuf *self) {
     pagerhist_clear(self);
     self->count = 0;
     self->start_of_data = 0;
+    for (size_t i = 1; i < self->num_segments; i++) free_segment(self->segments + i);
+    self->num_segments = 1;
 }
 
-static inline bool
+static bool
 pagerhist_write_bytes(PagerHistoryBuf *ph, const uint8_t *buf, size_t sz) {
     if (sz > ph->maximum_size) return false;
     if (!sz) return true;
@@ -192,11 +205,12 @@ pagerhist_write_bytes(PagerHistoryBuf *ph, const uint8_t *buf, size_t sz) {
     return true;
 }
 
-static inline bool
+static bool
 pagerhist_ensure_start_is_valid_utf8(PagerHistoryBuf *ph) {
     uint8_t scratch[8];
     size_t num = ringbuf_memcpy_from(scratch, ph->ringbuf, arraysz(scratch));
-    uint32_t state = UTF8_ACCEPT, codep;
+    uint32_t codep;
+    UTF8State state = UTF8_ACCEPT;
     size_t count = 0;
     size_t last_reject_at = 0;
     while (count < num) {
@@ -211,7 +225,7 @@ pagerhist_ensure_start_is_valid_utf8(PagerHistoryBuf *ph) {
     return false;
 }
 
-static inline bool
+static bool
 pagerhist_write_ucs4(PagerHistoryBuf *ph, const Py_UCS4 *buf, size_t sz) {
     uint8_t scratch[4];
     for (size_t i = 0; i < sz; i++) {
@@ -221,20 +235,20 @@ pagerhist_write_ucs4(PagerHistoryBuf *ph, const Py_UCS4 *buf, size_t sz) {
     return true;
 }
 
-static inline void
+static void
 pagerhist_push(HistoryBuf *self, ANSIBuf *as_ansi_buf) {
     PagerHistoryBuf *ph = self->pagerhist;
     if (!ph) return;
     const GPUCell *prev_cell = NULL;
     Line l = {.xnum=self->xnum};
     init_line(self, self->start_of_data, &l);
-    line_as_ansi(&l, as_ansi_buf, &prev_cell);
-    if (ringbuf_bytes_used(ph->ringbuf) && !l.continued) pagerhist_write_bytes(ph, (const uint8_t*)"\n", 1);
+    line_as_ansi(&l, as_ansi_buf, &prev_cell, 0, l.xnum, 0);
+    if (ringbuf_bytes_used(ph->ringbuf) && !l.attrs.continued) pagerhist_write_bytes(ph, (const uint8_t*)"\n", 1);
     pagerhist_write_bytes(ph, (const uint8_t*)"\x1b[m", 3);
     if (pagerhist_write_ucs4(ph, as_ansi_buf->buf, as_ansi_buf->len)) pagerhist_write_bytes(ph, (const uint8_t*)"\r", 1);
 }
 
-static inline index_type
+static index_type
 historybuf_push(HistoryBuf *self, ANSIBuf *as_ansi_buf) {
     index_type idx = (self->start_of_data + self->count) % self->ynum;
     init_line(self, idx, self->line);
@@ -249,7 +263,16 @@ void
 historybuf_add_line(HistoryBuf *self, const Line *line, ANSIBuf *as_ansi_buf) {
     index_type idx = historybuf_push(self, as_ansi_buf);
     copy_line(line, self->line);
-    *attrptr(self, idx) = (line->continued & CONTINUED_MASK) | (line->has_dirty_text ? TEXT_DIRTY_MASK : 0);
+    *attrptr(self, idx) = line->attrs;
+}
+
+bool
+historybuf_pop_line(HistoryBuf *self, Line *line) {
+    if (self->count <= 0) return false;
+    index_type idx = (self->start_of_data + self->count - 1) % self->ynum;
+    init_line(self, idx, line);
+    self->count--;
+    return true;
 }
 
 static PyObject*
@@ -269,7 +292,7 @@ __str__(HistoryBuf *self) {
     if (lines == NULL) return PyErr_NoMemory();
     for (index_type i = 0; i < self->count; i++) {
         init_line(self, index_of(self, i), self->line);
-        PyObject *t = line_as_unicode(self->line);
+        PyObject *t = line_as_unicode(self->line, false);
         if (t == NULL) { Py_CLEAR(lines); return NULL; }
         PyTuple_SET_ITEM(lines, i, t);
     }
@@ -299,10 +322,10 @@ as_ansi(HistoryBuf *self, PyObject *callback) {
     for(unsigned int i = 0; i < self->count; i++) {
         init_line(self, i, &l);
         if (i < self->count - 1) {
-            l.continued = *attrptr(self, index_of(self, i + 1)) & CONTINUED_MASK;
-        } else l.continued = false;
-        line_as_ansi(&l, &output, &prev_cell);
-        if (!l.continued) {
+            l.attrs.continued = attrptr(self, index_of(self, i + 1))->continued;
+        } else l.attrs.continued = false;
+        line_as_ansi(&l, &output, &prev_cell, 0, l.xnum, 0);
+        if (!l.attrs.continued) {
             ensure_space_for(&output, buf, Py_UCS4, output.len + 1, capacity, 2048, false);
             output.buf[output.len++] = 10; // 10 = \n
         }
@@ -319,12 +342,12 @@ end:
     Py_RETURN_NONE;
 }
 
-static inline Line*
+static Line*
 get_line(HistoryBuf *self, index_type y, Line *l) { init_line(self, index_of(self, self->count - y - 1), l); return l; }
 
-static inline char_type
+static char_type
 pagerhist_remove_char(PagerHistoryBuf *ph, unsigned *count, uint8_t record[8]) {
-    uint32_t codep, state = UTF8_ACCEPT;
+    uint32_t codep; UTF8State state = UTF8_ACCEPT;
     *count = 0;
     size_t num = ringbuf_bytes_used(ph->ringbuf);
     while (num--) {
@@ -341,11 +364,11 @@ static void
 pagerhist_rewrap_to(HistoryBuf *self, index_type cells_in_line) {
     PagerHistoryBuf *ph = self->pagerhist;
     if (!ph->ringbuf || !ringbuf_bytes_used(ph->ringbuf)) return;
-    PagerHistoryBuf *nph = PyMem_Calloc(sizeof(PagerHistoryBuf), 1);
+    PagerHistoryBuf *nph = calloc(sizeof(PagerHistoryBuf), 1);
     if (!nph) return;
     nph->maximum_size = ph->maximum_size;
     nph->ringbuf = ringbuf_new(MIN(ph->maximum_size, ringbuf_capacity(ph->ringbuf) + 4096));
-    if (!nph->ringbuf) { PyMem_Free(nph); return ; }
+    if (!nph->ringbuf) { free(nph); return ; }
     ssize_t ch_width = 0;
     unsigned count;
     uint8_t record[8];
@@ -395,22 +418,43 @@ pagerhist_write(HistoryBuf *self, PyObject *what) {
     Py_RETURN_NONE;
 }
 
+static const uint8_t*
+reverse_find(const uint8_t *haystack, size_t haystack_sz, const uint8_t *needle) {
+    const size_t needle_sz = strlen((const char*)needle);
+    if (!needle_sz || needle_sz > haystack_sz) return NULL;
+    const uint8_t *p = haystack + haystack_sz - (needle_sz - 1);
+    while (--p >= haystack) {
+        if (*p == needle[0] && memcmp(p, needle, MIN(needle_sz, haystack_sz - (p - haystack))) == 0) return p;
+    }
+    return NULL;
+}
+
 static PyObject*
-pagerhist_as_bytes(HistoryBuf *self, PyObject *args UNUSED) {
-    PagerHistoryBuf *ph = self->pagerhist;
+pagerhist_as_bytes(HistoryBuf *self, PyObject *args) {
+    int upto_output_start = 0;
+    if (!PyArg_ParseTuple(args, "|p", &upto_output_start)) return NULL;
+#define ph self->pagerhist
     if (!ph || !ringbuf_bytes_used(ph->ringbuf)) return PyBytes_FromStringAndSize("", 0);
     pagerhist_ensure_start_is_valid_utf8(ph);
     if (ph->rewrap_needed) pagerhist_rewrap_to(self, self->xnum);
 
     Line l = {.xnum=self->xnum}; get_line(self, 0, &l);
     size_t sz = ringbuf_bytes_used(ph->ringbuf);
-    if (!l.continued) sz += 1;
+    if (!l.attrs.continued) sz += 1;
     PyObject *ans = PyBytes_FromStringAndSize(NULL, sz);
     if (!ans) return NULL;
     uint8_t *buf = (uint8_t*)PyBytes_AS_STRING(ans);
     ringbuf_memcpy_from(buf, ph->ringbuf, sz);
-    if (!l.continued) buf[sz-1] = '\n';
+    if (!l.attrs.continued) buf[sz-1] = '\n';
+    if (upto_output_start) {
+        const uint8_t *p = reverse_find(buf, sz, (const uint8_t*)"\x1b]133;C\x1b\\");
+        if (p) {
+            PyObject *t = PyBytes_FromStringAndSize((const char*)p, sz - (p - buf));
+            Py_DECREF(ans); ans = t;
+        }
+    }
     return ans;
+#undef ph
 }
 
 static PyObject *
@@ -452,7 +496,7 @@ dirty_lines(HistoryBuf *self, PyObject *a UNUSED) {
 #define dirty_lines_doc "dirty_lines() -> Line numbers of all lines that have dirty text."
     PyObject *ans = PyList_New(0);
     for (index_type i = 0; i < self->count; i++) {
-        if (*attrptr(self, i) & TEXT_DIRTY_MASK) {
+        if (attrptr(self, i)->has_dirty_text) {
             PyList_Append(ans, PyLong_FromUnsignedLong(i));
         }
     }
@@ -477,8 +521,8 @@ static PyMethodDef methods[] = {
     METHOD(as_ansi, METH_O)
     METHODB(pagerhist_write, METH_O),
     METHODB(pagerhist_rewrap, METH_O),
-    METHODB(pagerhist_as_text, METH_NOARGS),
-    METHODB(pagerhist_as_bytes, METH_NOARGS),
+    METHODB(pagerhist_as_text, METH_VARARGS),
+    METHODB(pagerhist_as_bytes, METH_VARARGS),
     METHODB(as_text, METH_VARARGS),
     METHOD(dirty_lines, METH_NOARGS)
     METHOD(push, METH_VARARGS)
@@ -519,22 +563,23 @@ HistoryBuf *alloc_historybuf(unsigned int lines, unsigned int columns, unsigned
 
 #define init_src_line(src_y) init_line(src, map_src_index(src_y), src->line);
 
-#define is_src_line_continued(src_y) (map_src_index(src_y) < src->ynum - 1 ? (*attrptr(src, map_src_index(src_y + 1)) & CONTINUED_MASK) : false)
+#define is_src_line_continued(src_y) (map_src_index(src_y) < src->ynum - 1 ? (attrptr(src, map_src_index(src_y + 1))->continued) : false)
 
-#define next_dest_line(cont) *attrptr(dest, historybuf_push(dest, as_ansi_buf)) = cont & CONTINUED_MASK; dest->line->continued = cont;
+#define next_dest_line(cont) { LineAttrs *lap = attrptr(dest, historybuf_push(dest, as_ansi_buf)); *lap = src->line->attrs; if (cont) lap->continued = true; dest->line->attrs.continued = cont; }
 
 #define first_dest_line next_dest_line(false);
 
 #include "rewrap.h"
 
-void historybuf_rewrap(HistoryBuf *self, HistoryBuf *other, ANSIBuf *as_ansi_buf) {
+void
+historybuf_rewrap(HistoryBuf *self, HistoryBuf *other, ANSIBuf *as_ansi_buf) {
     while(other->num_segments < self->num_segments) add_segment(other);
     if (other->xnum == self->xnum && other->ynum == self->ynum) {
         // Fast path
         for (index_type i = 0; i < self->num_segments; i++) {
             memcpy(other->segments[i].cpu_cells, self->segments[i].cpu_cells, SEGMENT_SIZE * self->xnum * sizeof(CPUCell));
             memcpy(other->segments[i].gpu_cells, self->segments[i].gpu_cells, SEGMENT_SIZE * self->xnum * sizeof(GPUCell));
-            memcpy(other->segments[i].line_attrs, self->segments[i].line_attrs, SEGMENT_SIZE * sizeof(line_attrs_type));
+            memcpy(other->segments[i].line_attrs, self->segments[i].line_attrs, SEGMENT_SIZE * sizeof(LineAttrs));
         }
         other->count = self->count; other->start_of_data = self->start_of_data;
         return;
@@ -542,10 +587,9 @@ void historybuf_rewrap(HistoryBuf *self, HistoryBuf *other, ANSIBuf *as_ansi_buf
     if (other->pagerhist && other->xnum != self->xnum && ringbuf_bytes_used(other->pagerhist->ringbuf))
         other->pagerhist->rewrap_needed = true;
     other->count = 0; other->start_of_data = 0;
-    index_type x = 0, y = 0;
     if (self->count > 0) {
-        rewrap_inner(self, other, self->count, NULL, &x, &y, as_ansi_buf);
-        for (index_type i = 0; i < other->count; i++) *attrptr(other, (other->start_of_data + i) % other->ynum) |= TEXT_DIRTY_MASK;
+        rewrap_inner(self, other, self->count, NULL, NULL, as_ansi_buf);
+        for (index_type i = 0; i < other->count; i++) attrptr(other, (other->start_of_data + i) % other->ynum)->has_dirty_text = true;
     }
 }
 
