@@ -8,6 +8,7 @@
 #include "data-types.h"
 #include "screen.h"
 #include "monotonic.h"
+#include "window_logo.h"
 
 #define OPT(name) global_state.opts.name
 
@@ -23,17 +24,16 @@ typedef struct {
 typedef struct {
     monotonic_t visual_bell_duration, cursor_blink_interval, cursor_stop_blinking_after, mouse_hide_wait, click_interval;
     double wheel_scroll_multiplier, touch_scroll_multiplier;
+    int wheel_scroll_min_lines;
     bool enable_audio_bell;
     CursorShape cursor_shape;
     float cursor_beam_thickness;
     float cursor_underline_thickness;
-    unsigned int open_url_modifiers;
-    unsigned int rectangle_select_modifiers;
-    unsigned int terminal_select_modifiers;
     unsigned int url_style;
     unsigned int scrollback_pager_history_size;
-    char_type select_by_word_characters[256]; size_t select_by_word_characters_count;
-    color_type url_color, background, foreground, active_border_color, inactive_border_color, bell_border_color;
+    bool scrollback_fill_enlarged_window;
+    char_type *select_by_word_characters;
+    color_type url_color, background, foreground, active_border_color, inactive_border_color, bell_border_color, tab_bar_background, tab_bar_margin_color;
     color_type mark1_foreground, mark1_background, mark2_foreground, mark2_background, mark3_foreground, mark3_background;
     monotonic_t repaint_delay, input_delay;
     bool focus_follows_mouse;
@@ -42,14 +42,16 @@ typedef struct {
     unsigned int macos_option_as_alt;
     float macos_thicken_font;
     WindowTitleIn macos_show_window_title_in;
-    int adjust_line_height_px, adjust_column_width_px;
-    float adjust_line_height_frac, adjust_column_width_frac;
+    char *bell_path;
+    int adjust_line_height_px, adjust_column_width_px, adjust_baseline_px;
+    float adjust_line_height_frac, adjust_column_width_frac, adjust_baseline_frac;
     float background_opacity, dim_opacity;
 
-    char* background_image;
+    char *background_image, *default_window_logo;
     BackgroundImageLayout background_image_layout;
+    ImageAnchorPosition window_logo_position;
     bool background_image_linear;
-    float background_tint;
+    float background_tint, window_logo_alpha;
 
     bool dynamic_background_opacity;
     float inactive_text_alpha;
@@ -72,12 +74,27 @@ typedef struct {
         UrlPrefix *values;
         size_t num, max_prefix_len;
     } url_prefixes;
+    char_type *url_excluded_characters;
     bool detect_urls;
+    bool tab_bar_hidden;
+    double font_size;
+    struct {
+        double outer, inner;
+    } tab_bar_margin_height;
+    long macos_menubar_title_max_length;
 } Options;
 
+typedef struct WindowLogoRenderData {
+    window_logo_id_t id;
+    WindowLogo *instance;
+    ImageAnchorPosition position;
+    float alpha;
+    bool using_default;
+} WindowLogoRenderData;
+
 typedef struct {
     ssize_t vao_idx, gvao_idx;
-    float xstart, ystart, dx, dy;
+    float xstart, ystart, dx, dy, xratio, yratio;
     Screen *screen;
 } ScreenRenderData;
 
@@ -89,6 +106,7 @@ typedef struct {
     monotonic_t at;
     int button, modifiers;
     double x, y;
+    unsigned long num;
 } Click;
 
 #define CLICK_QUEUE_SZ 3
@@ -97,6 +115,12 @@ typedef struct {
     unsigned int length;
 } ClickQueue;
 
+typedef struct MousePosition {
+    unsigned int cell_x, cell_y;
+    double global_x, global_y;
+    bool in_left_half_of_cell;
+} MousePosition;
+
 typedef struct {
     id_type id;
     bool visible, cursor_visible_at_last_render;
@@ -104,21 +128,25 @@ typedef struct {
     CursorShape last_cursor_shape;
     PyObject *title;
     ScreenRenderData render_data;
-    struct {
-        unsigned int cell_x, cell_y;
-        double x, y;
-        bool in_left_half_of_cell;
-    } mouse_pos;
+    WindowLogoRenderData window_logo;
+    MousePosition mouse_pos;
     struct {
         unsigned int left, top, right, bottom;
     } padding;
     WindowGeometry geometry;
-    ClickQueue click_queue;
+    ClickQueue click_queues[8];
     monotonic_t last_drag_scroll_at;
+    uint32_t last_special_key_pressed;
+    struct {
+        unsigned width, height;
+        uint8_t *buf;
+        PyObject *last_drawn_title_object_id;
+    } title_bar_data;
 } Window;
 
 typedef struct {
-    uint32_t left, top, right, bottom, color;
+    float left, top, right, bottom;
+    uint32_t color;
 } BorderRect;
 
 typedef struct {
@@ -135,13 +163,6 @@ typedef struct {
     BorderRects border_rects;
 } Tab;
 
-#define MAX_KEY_COUNT 512
-
-typedef struct {
-    int x, y, w, h;
-    bool is_set;
-} OSWindowGeometry;
-
 enum RENDER_STATE { RENDER_FRAME_NOT_REQUESTED, RENDER_FRAME_REQUESTED, RENDER_FRAME_READY };
 typedef enum { NO_CLOSE_REQUESTED, CONFIRMABLE_CLOSE_REQUESTED, CLOSE_BEING_CONFIRMED, IMPERATIVE_CLOSE_REQUESTED } CloseRequest;
 
@@ -158,7 +179,10 @@ typedef struct {
     void *handle;
     id_type id;
     uint32_t offscreen_framebuffer;
-    OSWindowGeometry before_fullscreen;
+    struct {
+        int x, y, w, h;
+        bool is_set, was_maximized;
+    } before_fullscreen;
     int viewport_width, viewport_height, window_width, window_height;
     double viewport_x_ratio, viewport_y_ratio;
     Tab *tabs;
@@ -173,7 +197,7 @@ typedef struct {
     double logical_dpi_x, logical_dpi_y, font_sz_in_pts;
     bool mouse_button_pressed[32];
     PyObject *window_title;
-    bool is_key_pressed[MAX_KEY_COUNT];
+    bool disallow_title_changes, title_is_overriden;
     bool viewport_size_dirty, viewport_updated_at_least_once;
     LiveResizeInfo live_resize;
     bool has_pending_resizes, is_semi_transparent, shown_once, is_damaged;
@@ -195,7 +219,7 @@ typedef struct {
 typedef struct {
     Options opts;
 
-    id_type os_window_id_counter, tab_id_counter, window_id_counter;
+    id_type os_window_id_counter, tab_id_counter, window_id_counter, current_opengl_context_id;
     PyObject *boss;
     BackgroundImage *bgimage;
     OSWindow *os_windows;
@@ -206,11 +230,13 @@ typedef struct {
     bool debug_rendering, debug_font_fallback;
     bool has_pending_resizes, has_pending_closes;
     bool in_sequence_mode;
-    bool tab_bar_hidden;
-    double font_sz_in_pts;
+    bool check_for_active_animated_images;
     struct { double x, y; } default_dpi;
-    id_type active_drag_in_window;
+    id_type active_drag_in_window, tracked_drag_in_window;
+    int active_drag_button;
     CloseRequest quit_request;
+    bool redirect_mouse_handling;
+    WindowLogoTable *all_window_logos;
 } GlobalState;
 
 extern GlobalState global_state;
@@ -225,13 +251,16 @@ void gl_init(void);
 void remove_vao(ssize_t vao_idx);
 bool remove_os_window(id_type os_window_id);
 void make_os_window_context_current(OSWindow *w);
+void set_os_window_size(OSWindow *os_window, int x, int y);
+void get_os_window_size(OSWindow *os_window, int *w, int *h, int *fw, int *fh);
+void get_os_window_content_scale(OSWindow *os_window, double *xdpi, double *ydpi, float *xscale, float *yscale);
 void update_os_window_references(void);
 void mark_os_window_for_close(OSWindow* w, CloseRequest cr);
 void update_os_window_viewport(OSWindow *window, bool);
 bool should_os_window_be_rendered(OSWindow* w);
 void wakeup_main_loop(void);
 void swap_window_buffers(OSWindow *w);
-void make_window_context_current(OSWindow *w);
+bool make_window_context_current(id_type);
 void hide_mouse(OSWindow *w);
 bool is_mouse_hidden(OSWindow *w);
 void destroy_os_window(OSWindow *w);
@@ -247,7 +276,7 @@ ssize_t create_cell_vao(void);
 ssize_t create_graphics_vao(void);
 ssize_t create_border_vao(void);
 bool send_cell_data_to_gpu(ssize_t, ssize_t, float, float, float, float, Screen *, OSWindow *);
-void draw_cells(ssize_t, ssize_t, float, float, float, float, Screen *, OSWindow *, bool, bool);
+void draw_cells(ssize_t, ssize_t, const ScreenRenderData*, float, float, OSWindow *, bool, bool, Window*);
 void draw_centered_alpha_mask(OSWindow *w, size_t screen_width, size_t screen_height, size_t width, size_t height, uint8_t *canvas);
 void update_surface_size(int, int, uint32_t);
 void free_texture(uint32_t*);
@@ -256,16 +285,32 @@ void send_image_to_gpu(uint32_t*, const void*, int32_t, int32_t, bool, bool, boo
 void send_sprite_to_gpu(FONTS_DATA_HANDLE fg, unsigned int, unsigned int, unsigned int, pixel*);
 void blank_canvas(float, color_type);
 void blank_os_window(OSWindow *);
-void set_titlebar_color(OSWindow *w, color_type color);
+void set_titlebar_color(OSWindow *w, color_type color, bool use_system_color, unsigned int system_color);
 FONTS_DATA_HANDLE load_fonts_data(double, double, double);
 void send_prerendered_sprites_for_window(OSWindow *w);
 #ifdef __APPLE__
-void get_cocoa_key_equivalent(int, int, char *key, size_t key_sz, int*);
+void get_cocoa_key_equivalent(uint32_t, int, char *key, size_t key_sz, int*);
 typedef enum {
-    PREFERENCES_WINDOW = 1,
-    NEW_OS_WINDOW = 2,
-    NEW_OS_WINDOW_WITH_WD = 4,
-    NEW_TAB_WITH_WD = 8
+    PREFERENCES_WINDOW,
+    NEW_OS_WINDOW,
+    NEW_OS_WINDOW_WITH_WD,
+    NEW_TAB_WITH_WD,
+    CLOSE_OS_WINDOW,
+    CLOSE_TAB,
+    NEW_TAB,
+    NEXT_TAB,
+    PREVIOUS_TAB,
+    DETACH_TAB,
+    LAUNCH_URLS,
+    NEW_WINDOW,
+    CLOSE_WINDOW,
+    RESET_TERMINAL,
+    CLEAR_TERMINAL_AND_SCROLLBACK,
+    RELOAD_CONFIG,
+    TOGGLE_MACOS_SECURE_KEYBOARD_ENTRY,
+    TOGGLE_FULLSCREEN,
+
+    NUM_COCOA_PENDING_ACTIONS
 } CocoaPendingAction;
 void set_cocoa_pending_action(CocoaPendingAction action, const char*);
 #endif
@@ -282,3 +327,21 @@ void os_window_update_size_increments(OSWindow *window);
 void set_os_window_title_from_window(Window *w, OSWindow *os_window);
 void update_os_window_title(OSWindow *os_window);
 void fake_scroll(Window *w, int amount, bool upwards);
+Window* window_for_window_id(id_type kitty_window_id);
+bool mouse_open_url(Window *w);
+bool mouse_set_last_visited_cmd_output(Window *w);
+bool mouse_select_cmd_output(Window *w);
+bool move_cursor_to_mouse_if_at_shell_prompt(Window *w);
+void mouse_selection(Window *w, int code, int button);
+const char* format_mods(unsigned mods);
+void send_pending_click_to_window_id(id_type, void*);
+void send_pending_click_to_window(Window*, void*);
+void get_platform_dependent_config_values(void *glfw_window);
+bool draw_window_title(OSWindow *window, const char *text, color_type fg, color_type bg, uint8_t *output_buf, size_t width, size_t height);
+uint8_t* draw_single_ascii_char(const char ch, size_t *result_width, size_t *result_height);
+bool is_os_window_fullscreen(OSWindow *);
+void update_ime_focus(OSWindow* osw, bool focused);
+void update_ime_position(Window* w, Screen *screen);
+bool update_ime_position_for_window(id_type window_id, bool force, bool lost_focus);
+void set_ignore_os_keyboard_processing(bool enabled);
+void update_menu_bar_title(PyObject *title UNUSED);
