@@ -1,5 +1,4 @@
 #!/usr/bin/env python3
-# vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
 import os
@@ -7,8 +6,8 @@ import tempfile
 
 from kitty.config import build_ansi_color_table, defaults
 from kitty.fast_data_types import (
-    REVERSE, ColorProfile, Cursor as C, HistoryBuf, LineBuf,
-    parse_input_from_terminal, truncate_point_for_length, wcswidth, wcwidth
+    ColorProfile, Cursor as C, HistoryBuf, LineBuf, Color,
+    parse_input_from_terminal, truncate_point_for_length, wcswidth, wcwidth, strip_csi
 )
 from kitty.rgb import to_color
 from kitty.utils import is_path_in_temp_dir, sanitize_title
@@ -33,8 +32,12 @@ class TestDataTypes(BaseTest):
         for x in 'xxx #12 #1234 rgb:a/b'.split():
             self.assertIsNone(to_color(x))
 
-        def c(spec, r=0, g=0, b=0):
-            self.ae(tuple(to_color(spec)), (r, g, b))
+        def c(spec, r=0, g=0, b=0, a=0):
+            c = to_color(spec)
+            self.ae(c.red, r)
+            self.ae(c.green, g)
+            self.ae(c.blue, b)
+            self.ae(c.alpha, a)
 
         c('#eee', 0xee, 0xee, 0xee)
         c('#234567', 0x23, 0x45, 0x67)
@@ -43,6 +46,15 @@ class TestDataTypes(BaseTest):
         c('rgb:23/45/67', 0x23, 0x45, 0x67)
         c('rgb:abc/abc/def', 0xab, 0xab, 0xde)
         c('red', 0xff)
+        self.ae(int(Color(1, 2, 3)), 0x10203)
+        base = Color(12, 12, 12)
+        a = Color(23, 23, 23)
+        b = Color(100, 100, 100)
+        self.assertLess(base.contrast(a), base.contrast(b))
+        self.ae(Color(1, 2, 3).as_sgr, ':2:1:2:3')
+        self.ae(Color(1, 2, 3).as_sharp, '#010203')
+        self.ae(Color(1, 2, 3, 4).as_sharp, '#04010203')
+        self.ae(Color(1, 2, 3, 4).rgb, 0x10203)
 
     def test_linebuf(self):
         old = filled_line_buf(2, 3, filled_cursor())
@@ -51,7 +63,7 @@ class TestDataTypes(BaseTest):
         self.ae(new.line(0), old.line(1))
         new.clear()
         self.ae(str(new.line(0)), '')
-        old.set_attribute(REVERSE, False)
+        old.set_attribute('reverse', False)
         for y in range(old.ynum):
             for x in range(old.xnum):
                 l0 = old.line(y)
@@ -243,13 +255,16 @@ class TestDataTypes(BaseTest):
         l0 = create('file:///etc/test')
         self.ae(l0.url_start_at(0), 0)
 
-        for trail in '.,]>)\\':
+        for trail in '.,\\':
             lx = create("http://xyz.com" + trail)
             self.ae(lx.url_end_at(0), len(lx) - 2)
+        for trail in ')}]>':
+            lx = create("http://xyz.com" + trail)
+            self.ae(lx.url_end_at(0), len(lx) - 1)
         l0 = create("ftp://abc/")
         self.ae(l0.url_end_at(0), len(l0) - 1)
         l2 = create("http://-abcd] ")
-        self.ae(l2.url_end_at(0), len(l2) - 3)
+        self.ae(l2.url_end_at(0), len(l2) - 2)
         l3 = create("http://ab.de           ")
         self.ae(l3.url_start_at(4), 0)
         self.ae(l3.url_start_at(5), 0)
@@ -366,9 +381,12 @@ class TestDataTypes(BaseTest):
         self.ae(wcswidth('\U0001F1E6a\U0001F1E8a'), 6)
         self.ae(wcswidth('\U0001F1E6\U0001F1E8a'), 3)
         self.ae(wcswidth('\U0001F1E6\U0001F1E8\U0001F1E6'), 4)
+        self.ae(wcswidth('a\u00adb'), 2)
         # Regional indicator symbols (unicode flags) are defined as having
-        # Emoji_Presentation so must have width 2
+        # Emoji_Presentation so must have width 2 but combined must have
+        # width 2 not 4
         self.ae(tuple(map(w, '\U0001f1ee\U0001f1f3')), (2, 2))
+        self.ae(wcswidth('\U0001f1ee\U0001f1f3'), 2)
         tpl = truncate_point_for_length
         self.ae(tpl('abc', 4), 3)
         self.ae(tpl('abc', 2), 2)
@@ -419,9 +437,9 @@ class TestDataTypes(BaseTest):
         c = ColorProfile()
         c.update_ansi_color_table(build_ansi_color_table())
         for i in range(8):
-            col = getattr(defaults, 'color{}'.format(i))
-            self.assertEqual(c.as_color(i << 8 | 1), (col[0], col[1], col[2]))
-        self.ae(c.as_color(255 << 8 | 1), (0xee, 0xee, 0xee))
+            col = getattr(defaults, f'color{i}')
+            self.assertEqual(c.as_color(i << 8 | 1), col)
+        self.ae(c.as_color(255 << 8 | 1), Color(0xee, 0xee, 0xee))
 
     def test_historybuf(self):
         lb = filled_line_buf()
@@ -504,3 +522,11 @@ class TestDataTypes(BaseTest):
         a = []
         hb.as_ansi(a.append)
         self.ae(a, [str(hb.line(i)) + '\n' for i in range(hb.count - 1, -1, -1)])
+
+    def test_strip_csi(self):
+        def q(x, y=''):
+            self.ae(y or x, strip_csi(x))
+        q('test')
+        q('a\x1bbc', 'ac')
+        q('a\x1b[bc', 'ac')
+        q('a\x1b[12;34:43mbc', 'abc')
