@@ -1,11 +1,11 @@
 #!/usr/bin/env python3
-# vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
 from kitty.fast_data_types import (
     DECAWM, DECCOLM, DECOM, IRM, Cursor, parse_bytes
 )
 from kitty.marks import marker_from_function, marker_from_regex
+from kitty.window import pagerhist
 
 from . import BaseTest
 
@@ -93,6 +93,17 @@ class TestScreen(BaseTest):
         self.ae(str(s.line(4)), 'a\u0306b1\u030623')
         self.ae((s.cursor.x, s.cursor.y), (2, 4))
 
+    def test_rep(self):
+        s = self.create_screen()
+        s.draw('a')
+        parse_bytes(s, b'\x1b[b')
+        self.ae(str(s.line(0)), 'aa')
+        parse_bytes(s, b'\x1b[3b')
+        self.ae(str(s.line(0)), 'a'*5)
+        s.draw(' ')
+        parse_bytes(s, b'\x1b[3b')
+        self.ae(str(s.line(1)), ' '*4)
+
     def test_emoji_skin_tone_modifiers(self):
         s = self.create_screen()
         q = chr(0x1f469) + chr(0x1f3fd)
@@ -100,12 +111,29 @@ class TestScreen(BaseTest):
         self.ae(str(s.line(0)), q)
         self.ae(s.cursor.x, 2)
 
+    def test_regional_indicators(self):
+        s = self.create_screen()
+        flag = '\U0001f1ee\U0001f1f3'
+        s.draw(flag)
+        self.ae(str(s.line(0)), flag)
+        self.ae(s.cursor.x, 2)
+        s = self.create_screen()
+        s.draw('a'), s.draw(flag[0]), s.draw('b')
+        self.ae(str(s.line(0)), 'a' + flag[0] + 'b')
+        self.ae(s.cursor.x, 4)
+
     def test_zwj(self):
         s = self.create_screen(cols=20)
         q = '\U0001f468\u200d\U0001f469\u200d\U0001f467\u200d\U0001f466'
         s.draw(q)
         self.ae(q, str(s.line(0)))
         self.ae(s.cursor.x, 8)
+        for x in '\u200b\u200c\u200d':
+            s = self.create_screen()
+            q = f'X{x}Y'
+            s.draw(q)
+            self.ae(q, str(s.line(0)))
+            self.ae(s.cursor.x, 2)
 
     def test_char_manipulation(self):
         s = self.create_screen()
@@ -292,6 +320,65 @@ class TestScreen(BaseTest):
         s.resize(s.lines - 1, s.columns)
         self.ae(x_before, s.cursor.x)
 
+    def test_scrollback_fill_after_resize(self):
+        def prepare_screen(content=()):
+            ans = self.create_screen(options={'scrollback_fill_enlarged_window': True})
+            for line in content:
+                ans.draw(line)
+                ans.linefeed()
+                ans.carriage_return()
+            return ans
+
+        def assert_lines(*lines):
+            return self.ae(lines, tuple(str(s.line(i)) for i in range(s.lines)))
+
+        # test the reverse scroll function
+        s = prepare_screen(map(str, range(6)))
+        assert_lines('2', '3', '4', '5', '')
+        s.reverse_scroll(2, True)
+        assert_lines('0', '1', '2', '3', '4')
+
+        # Height increased, width unchanged → pull down lines to fill new space at the top
+        s = prepare_screen(map(str, range(6)))
+        assert_lines('2', '3', '4', '5', '')
+        dist_from_bottom = s.lines - s.cursor.y
+        s.resize(7, s.columns)
+        assert_lines('0', '1', '2', '3', '4', '5', '')
+        self.ae(dist_from_bottom, s.lines - s.cursor.y)
+
+        # Height increased, width increased → rewrap, pull down
+        s = prepare_screen(['0', '1', '2', '3' * 15])
+        assert_lines('2', '33333', '33333', '33333', '')
+        s.resize(7, 12)
+        assert_lines('0', '1', '2', '333333333333', '333', '', '')
+
+        # Height increased, width decreased → rewrap, pull down if possible
+        s = prepare_screen(['0', '1', '2', '3' * 5])
+        assert_lines('0', '1', '2', '33333', '')
+        s.resize(6, 4)
+        assert_lines('0', '1', '2', '3333', '3', '')
+
+        # Height unchanged, width increased → rewrap, pull down if possible
+        s = prepare_screen(['0', '1', '2', '3' * 15])
+        assert_lines('2', '33333', '33333', '33333', '')
+        s.resize(s.lines, 12)
+        assert_lines('1', '2', '333333333333', '333', '')
+
+        # Height decreased, width increased → rewrap, pull down if possible
+        s = prepare_screen(['0', '1', '2', '3' * 15])
+        assert_lines('2', '33333', '33333', '33333', '')
+        s.resize(4, 12)
+        assert_lines('2', '333333333333', '333', '')
+
+        # Height increased with large continued text
+        s = self.create_screen(options={'scrollback_fill_enlarged_window': True})
+        s.draw(('x' * (s.columns * s.lines * 2)) + 'abcde')
+        s.carriage_return(), s.linefeed()
+        s.draw('>')
+        assert_lines('xxxxx', 'xxxxx', 'xxxxx', 'abcde', '>')
+        s.resize(s.lines + 2, s.columns)
+        assert_lines('xxxxx', 'xxxxx', 'xxxxx', 'xxxxx', 'xxxxx', 'abcde', '>')
+
     def test_tab_stops(self):
         # Taken from vttest/main.c
         s = self.create_screen(cols=80, lines=2)
@@ -443,6 +530,37 @@ class TestScreen(BaseTest):
         self.ae(s.text_for_selection(), expected)
         s.scroll(2, True)
         self.ae(s.text_for_selection(), expected)
+        s.reset()
+        s.draw('ab   cd')
+        s.start_selection(0, 0)
+        s.update_selection(1, 3)
+        self.ae(s.text_for_selection(), ('ab   ', 'cd'))
+        self.ae(s.text_for_selection(False, True), ('ab', 'cd'))
+        s.reset()
+        s.draw('ab        cd')
+        s.start_selection(0, 0)
+        s.update_selection(3, 4)
+        self.ae(s.text_for_selection(), ('ab   ', '     ', 'cd'))
+        self.ae(s.text_for_selection(False, True), ('ab', '\n', 'cd'))
+        s.reset()
+        s.draw('a')
+        s.select_graphic_rendition(32)
+        s.draw('b')
+        s.select_graphic_rendition(39)
+        s.draw('c  xy')
+        s.start_selection(0, 0)
+        s.update_selection(1, 3)
+        self.ae(s.text_for_selection(), ('abc  ', 'xy'))
+        self.ae(s.text_for_selection(True), ('a\x1b[32mb\x1b[39mc  ', 'xy', '\x1b[m'))
+        self.ae(s.text_for_selection(True, True), ('a\x1b[32mb\x1b[39mc', 'xy', '\x1b[m'))
+
+    def test_soft_hyphen(self):
+        s = self.create_screen()
+        s.draw('a\u00adb')
+        self.ae(s.cursor.x, 2)
+        s.start_selection(0, 0)
+        s.update_selection(2, 0)
+        self.ae(s.text_for_selection(), ('a\u00adb',))
 
     def test_variation_selectors(self):
         s = self.create_screen()
@@ -592,6 +710,13 @@ class TestScreen(BaseTest):
         self.ae(s.marked_cells(), cells(8))
         s.set_marker(marker_from_regex('\t', 3))
         self.ae(s.marked_cells(), cells(*range(8)))
+        s = self.create_screen()
+        s.cursor.x = 2
+        s.draw('x')
+        s.cursor.x += 1
+        s.draw('x')
+        s.set_marker(marker_from_function(mark_x))
+        self.ae(s.marked_cells(), [(2, 0, 1), (4, 0, 2)])
 
     def test_hyperlinks(self):
         s = self.create_screen()
@@ -716,3 +841,248 @@ class TestScreen(BaseTest):
 
         self.ae(str(s.linebuf), '0\n5\n6\n7\n\n')
         self.ae(str(s.historybuf), '')
+
+    def test_osc_52(self):
+        s = self.create_screen()
+        c = s.callbacks
+
+        def send(what: str):
+            return parse_bytes(s, f'\033]52;p;{what}\a'.encode('ascii'))
+
+        def t(q, use_pending_mode, *expected):
+            c.clear()
+            if use_pending_mode:
+                parse_bytes(s, b'\033[?2026h')
+            send(q)
+            if use_pending_mode:
+                self.ae(c.cc_buf, [])
+                parse_bytes(s, b'\033[?2026l')
+            self.ae(c.cc_buf, list(expected))
+
+        for use_pending_mode in (False, True):
+            t('XYZ', use_pending_mode, ('p;XYZ', False))
+            t('a' * 8192, use_pending_mode, ('p;' + 'a' * (8192 - 6), True), (';' + 'a' * 6, False))
+            t('', use_pending_mode, ('p;', False))
+            t('!', use_pending_mode, ('p;!', False))
+
+    def test_key_encoding_flags_stack(self):
+        s = self.create_screen()
+        c = s.callbacks
+
+        def w(code, p1='', p2=''):
+            p = f'{p1}'
+            if p2:
+                p += f';{p2}'
+            return parse_bytes(s, f'\033[{code}{p}u'.encode('ascii'))
+
+        def ac(flags):
+            parse_bytes(s, b'\033[?u')
+            self.ae(c.wtcbuf, f'\033[?{flags}u'.encode('ascii'))
+            c.clear()
+
+        ac(0)
+        w('=', 0b1001)
+        ac(0b1001)
+        w('=', 0b0011, 2)
+        ac(0b1011)
+        w('=', 0b0110, 3)
+        ac(0b1001)
+        s.reset()
+        ac(0)
+
+        w('>', 0b0011)
+        ac(0b0011)
+        w('=', 0b1111)
+        ac(0b1111)
+        w('>', 0b10)
+        ac(0b10)
+        w('<')
+        ac(0b1111)
+        for i in range(10):
+            w('<')
+            ac(0)
+        s.reset()
+
+        for i in range(1, 16):
+            w('>', i)
+        ac(15)
+        w('<'), ac(14), w('<'), ac(13)
+
+    def test_color_stack(self):
+        s = self.create_screen()
+        c = s.callbacks
+
+        def w(code):
+            return parse_bytes(s, ('\033[' + code).encode('ascii'))
+
+        def ac(idx, count):
+            self.ae(c.wtcbuf, f'\033[{idx};{count}#Q'.encode('ascii'))
+            c.clear()
+
+        w('#R')
+        ac(0, 0)
+
+        w('#P')
+        w('#R')
+        ac(0, 1)
+        w('#10P')
+        w('#R')
+        ac(0, 1)
+        w('#Q')
+        w('#R')
+        ac(0, 0)
+        for i in range(20):
+            w('#P')
+        w('#R')
+        ac(9, 10)
+
+    def test_detect_url(self):
+        s = self.create_screen(cols=30)
+
+        def ae(expected, x=3, y=0):
+            s.detect_url(x, y)
+            url = ''.join(s.text_for_marked_url())
+            self.assertEqual(expected, url)
+
+        def t(url, x=0, y=0, before='', after=''):
+            s.reset()
+            s.cursor.x = x
+            s.cursor.y = y
+            s.draw(before + url + after)
+            ae(url, x=x + 1 + len(before), y=y)
+
+        t('http://moo.com')
+        t('http://moo.com/something?else=+&what-')
+        for (st, e) in '() {} [] <>'.split():
+            t('http://moo.com', before=st, after=e)
+        for trailer in ')-=]}':
+            t('http://moo.com' + trailer)
+        for trailer in '{([':
+            t('http://moo.com', after=trailer)
+        t('http://moo.com', x=s.columns - 9)
+
+    def test_prompt_marking(self):
+        s = self.create_screen()
+
+        def mark_prompt():
+            parse_bytes(s, b'\033]133;A\007')
+
+        def mark_output():
+            parse_bytes(s, b'\033]133;C\007')
+
+        for i in range(4):
+            mark_prompt()
+            s.draw(f'$ {i}')
+            s.carriage_return()
+            s.index(), s.index()
+        self.ae(s.scrolled_by, 0)
+        self.assertTrue(s.scroll_to_prompt())
+        self.ae(str(s.visual_line(0)), '$ 1')
+        self.assertTrue(s.scroll_to_prompt())
+        self.ae(str(s.visual_line(0)), '$ 0')
+        self.assertFalse(s.scroll_to_prompt())
+        self.assertTrue(s.scroll_to_prompt(1))
+        self.ae(str(s.visual_line(0)), '$ 1')
+        self.assertTrue(s.scroll_to_prompt(1))
+        self.ae(str(s.visual_line(0)), '$ 2')
+        self.assertFalse(s.scroll_to_prompt(1))
+
+        s = self.create_screen()
+        mark_prompt(), s.draw('$ 0')
+        s.carriage_return(), s.index()
+        mark_prompt(), s.draw('$ 1')
+        for i in range(s.lines):
+            s.carriage_return(), s.index()
+            s.draw(str(i))
+        self.assertTrue(s.scroll_to_prompt())
+        self.ae(str(s.visual_line(0)), '$ 1')
+
+        def lco(as_ansi=False):
+            a = []
+            if s.cmd_output(0, a.append, as_ansi):
+                pht = pagerhist(s, as_ansi=as_ansi, upto_output_start=True)
+                if pht:
+                    a.insert(0, pht)
+            return ''.join(a)
+
+        def fco():
+            a = []
+            s.cmd_output(1, a.append)
+            return ''.join(a)
+
+        def lvco():
+            a = []
+            s.cmd_output(2, a.append)
+            return ''.join(a)
+
+        s = self.create_screen()
+        s.draw('abcd'), s.index(), s.carriage_return()
+        s.draw('12'), s.index(), s.carriage_return()
+        self.ae(fco(), '')
+        self.ae(lco(), 'abcd\n12')
+        s = self.create_screen()
+        mark_prompt(), s.draw('$ 0')
+        s.carriage_return(), s.index()
+        mark_output()
+        s.draw('abcd'), s.index(), s.carriage_return()
+        s.draw('12'), s.index(), s.carriage_return()
+        mark_prompt(), s.draw('$ 1')
+        self.ae(fco(), 'abcd\n12')
+        self.ae(lco(), 'abcd\n12')
+        self.ae(lco(as_ansi=True), '\x1b[m\x1b]133;C\x1b\\abcd\n\x1b[m12')
+
+        def draw_prompt(x):
+            mark_prompt(), s.draw(f'$ {x}'), s.carriage_return(), s.index()
+
+        def draw_output(n, x='', m=True):
+            if m:
+                mark_output()
+            for i in range(n):
+                s.draw(f'{i}{x}'), s.index(), s.carriage_return()
+
+        s = self.create_screen(cols=5, lines=5, scrollback=15)
+        draw_output(1, 'start', False)
+        draw_prompt('0'), draw_output(3)
+        draw_prompt('1')
+        draw_prompt('2'), draw_output(2, 'x')
+
+        # last cmd output
+        # test: find upwards
+        self.ae(s.scrolled_by, 0)
+        self.ae(lco(), '0x\n1x')
+        # get output after scroll up
+        s.scroll_to_prompt()
+        self.ae(s.scrolled_by, 4)
+        self.ae(str(s.visual_line(0)), '$ 0')
+        self.ae(lco(), '0x\n1x')
+
+        # first cmd output on screen
+        # test: find around
+        self.ae(fco(), '0\n1\n2')
+        s.scroll(2, False)
+        self.ae(s.scrolled_by, 2)
+        self.ae(str(s.visual_line(0)), '1')
+        self.ae(fco(), '0x\n1x')
+        # test: find downwards
+        s.scroll(2, False)
+        self.ae(str(s.visual_line(0)), '$ 1')
+        self.ae(fco(), '0x\n1x')
+
+        # resize
+        # get last cmd output with continued output mark
+        draw_prompt('3'), draw_output(1, 'long_line'), draw_output(2, 'l', False)
+        s.resize(4, 5)
+        s.scroll_to_prompt(-4)
+        self.ae(str(s.visual_line(0)), '$ 0')
+        self.ae(lco(), '0long_line\n0l\n1l')
+
+        # last visited cmd output
+        self.ae(lvco(), '0\n1\n2')
+        s.scroll_to_prompt(1)
+        self.ae(lvco(), '0x\n1x')
+
+        # last command output from pager history
+        s = self.create_screen()
+        draw_prompt('p1')
+        draw_output(30)
+        self.ae(tuple(map(int, lco().split())), tuple(range(0, 30)))
